"""
Example for usage of the Turbo Design Tool
"""
import carbatpy as cb
import sys
import time
import datetime
from carbatpy.optional.spp_machines import Compressor, Turbine, require_spp_machines

#%%
# Compressor
# Initiate Compressor
print(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))

# Test compressor runtime
start = time.time()
Fluid= "pentane * propane * ethane"
Composition = [0.11, 0.7, 0.19]
P_in = 128604.9122629143
Temp_in = 286.5733615490751
P_out = 3361353.5072705583
V_dot_in = 17.50971555097697

Results1 = Compressor()
OutletState, MachineData = Results1.designCompressor(Fluid, 
                                                    Composition, 
                                                    P_in, 
                                                    Temp_in, 
                                                    P_out, 
                                                    V_dot_in, 
                                                    GearedCompressor=False,
                                                    Plot=False,
                                                    EfficiencyConvergence=5e-2,
                                                    MaxIterations=10
                                                    )
print(Results1.BasicParameters.StageNumber)
end = time.time()
runtime1 = end - start
print(runtime1/60)

sys.exit()

start = time.time()
Fluid= "hexane * pentane * propane * ethane"
Composition = [0.03118315, 0.12699793, 0.82770259, 0.01411633]
P_in = 128604.9122629143
Temp_in = 286.5733615490751
P_out = 3361353.5072705583
V_dot_in = 17.50971555097697

Results2 = Compressor()
OutletState, MachineData = Results2.designCompressor(Fluid, 
                                                    Composition, 
                                                    P_in, 
                                                    Temp_in, 
                                                    P_out, 
                                                    V_dot_in, 
                                                    GearedCompressor=False,
                                                    Plot=False,
                                                    EfficiencyConvergence=5e-2,
                                                    MaxIterations=10
                                                    )
print(Results2.BasicParameters.StageNumber)
end = time.time()
runtime1 = end - start
print(runtime1/60)

start = time.time()
Composition = [0.05, 0.05, 0.75, 0.15]

Results3 = Compressor()
OutletState, MachineData = Results3.designCompressor(Fluid, 
                                                    Composition, 
                                                    P_in, 
                                                    Temp_in, 
                                                    P_out, 
                                                    V_dot_in, 
                                                    GearedCompressor=False,
                                                    Plot=False,
                                                    EfficiencyConvergence=5e-2,
                                                    MaxIterations=10
                                                    )

print(Results3.BasicParameters.StageNumber)
end = time.time()
runtime1 = end - start
print(runtime1/60)

start = time.time()
Fluid = "propane"
Composition = [1.0]
Results4 = Compressor()
OutletState, MachineData = Results4.designCompressor(Fluid, 
                                                    Composition, 
                                                    P_in, 
                                                    Temp_in, 
                                                    P_out, 
                                                    V_dot_in, 
                                                    GearedCompressor=False,
                                                    Plot=False,
                                                    EfficiencyConvergence=5e-2,
                                                    MaxIterations=10
                                                    )
print(Results4.BasicParameters.StageNumber)
end = time.time()
runtime4 = end - start
print(runtime4/60)

sys.exit()

#%%
Results = Compressor()
# Example from Repository
# Create inputs for the Tool
InletTemperature = 300           # Inlet temperature in K
InletPressure = 1e5              # Inlet pressure in Pa
OutletPressure = 30e5             # Desired outlet pressure in Pa
WorkingFluid = "ethane * propane * pentane * hexane"  # Working fluid as named in REFPROP
WorkingFluidComposition = [0.15, 0.75, 0.05, 0.05]    # Working fluid composition in mole fractions
InletVolumeFlow = 1.5              # Inlet volume flow in m^3/s

# Calling the design method
OutletState, MachineData = Results.designCompressor(WorkingFluid, 
                                                    WorkingFluidComposition, 
                                                    InletPressure, 
                                                    InletTemperature, 
                                                    OutletPressure, 
                                                    InletVolumeFlow, 
                                                    GearedCompressor=False,
                                                    Plot=True,
                                                    EfficiencyConvergence=5e-2
                                                    )

# Results:
    # OutletState: working fluid state at the compressor outlet as a carbatpy state object
    # MachineData: dictionary containing the details of the compressor design, used by further methods
#%%
# Outletstate
print(OutletState.state)

# Total efficiency in MachineData at the last stage:
print(Results.IsentropicEfficiency[-1])
print(Results.PolytropicEfficiency)

#%%
# Turbine
# Create inputs for the Tool
InletTemperature = 300+273.15
InletPressure = 20e5
OutletPressure = 1e5 
WorkingFluid = "ethane * propane * pentane * hexane"
WorkingFluidComposition = [0.15, 0.75, 0.05, 0.05]
MassFlow = 1
TS = cb.CB_DEFAULTS['Fluid_Defaults']['TRANS_STRING']

myfluid = cb.init_fluid(WorkingFluid, WorkingFluidComposition)
InletState = myfluid.set_state([InletTemperature, InletPressure], "TP", 
                               wanted=TS,output="FluidState")

# Class initialisation
Results_turbine = Turbine()

# Calling the design method
OutletState_turbine, MachineData_turbine = Results_turbine.designTurbine(WorkingFluid, 
                                                 WorkingFluidComposition, 
                                                 InletState, 
                                                 OutletPressure, 
                                                 MassFlow, 
                                                 Plot=True)

#%%
# Outletstate
print(OutletState_turbine.state)

# Total efficiency in MachineData at the last stage:
print(Results_turbine.IsentropicEfficiency[-1])
print(Results_turbine.PolytropicEfficiency)



# -*- coding: utf-8 -*-
"""
Heat pump with two two-tank storages (component-based formulation, comp.py).

Created on Sun Jan  4 11:06:10 2026

Author: atakan
University of Duisburg-Essen, Germany

Within the Priority Programme: "Carnot Batteries: Inverse Design from
Markets to Molecules" (SPP 2403)
https://www.uni-due.de/spp2403/
https://git.uni-due.de/spp-2403/residuals_weather_storage
"""
import copy
from types import SimpleNamespace
import numpy as np
import matplotlib.pyplot as plt
import carbatpy as cb
# import pandas as pd  # only needed in the optional optimization example below


def deep_merge(dst: dict, src: dict) -> dict:
    """
    Recursively merge two nested dictionaries.

    - Values from src are written into dst.
    - For nested dicts, merging is done recursively.
    - Non-dict values overwrite the value in dst.

    Notes
    -----
    - dst is modified in-place.
    - src is not modified.

    Parameters
    ----------
    dst : dict
        Target dictionary; will be updated in-place.
    src : dict
        Source dictionary; its values are merged into dst.

    Returns
    -------
    dict
        The updated dst dictionary (same reference).
    """
    for k, v in src.items():
        if isinstance(v, dict) and isinstance(dst.get(k), dict):
            deep_merge(dst[k], v)
        else:
            dst[k] = v
    return dst


def resolve_config(dir_or_dict, override=None) -> dict:
    """
    Create an isolated configuration and apply optional overrides.

    - If dir_or_dict is a path, load the YAML configuration.
    - If dir_or_dict is a dict, use a deep copy of it.
    - Overrides are deep-merged into the base configuration.
    - Always returns a new, isolated structure (no references to global defaults).

    Parameters
    ----------
    dir_or_dict : str or dict
        Path to a YAML file or a configuration dictionary.
    override : dict, optional
        Dictionary with parameters to override (e.g., for optimization).

    Returns
    -------
    dict
        Fully isolated and merged configuration structure.
    """
    if isinstance(dir_or_dict, dict):
        base = copy.deepcopy(dir_or_dict)
    else:
        # read_config gibt direkt zurück wenn dict, daher hier deepcopy
        loaded = cb.utils.io_utils.read_config(dir_or_dict)
        base = copy.deepcopy(loaded)
    
    if override:
        deep_merge(base, copy.deepcopy(override))
    return base


def heat_pump(dir_name, **kwargs):
    """
    Compute a heat pump with two two-tank storages (component-based formulation).

    The function builds an isolated, merged configuration from a base (file or dict),
    computes the component sequence (start, compressor, condenser, throttle, evaporator),
    and returns COP, outputs, warnings, and optionally a combined plot.

    Key properties
    - No in-place modifications to CB_DEFAULTS or externally provided configurations.
    - Configurations are handled in isolation (deepcopy/recursive merges).
    - Matplotlib Figure/Axes are consistently passed to all plot calls so that the
      entire cycle is drawn on the same axes.

    Parameters
    ----------
    dir_name : str or dict
        Path to a YAML configuration file or a configuration dict.
    config : dict, optional (via kwargs)
        Overrides to the base configuration (e.g., for optimization).
    verbose : bool, optional (via kwargs)
        If True, print additional information.
    plotting : bool, optional (via kwargs)
        If True, create a combined T-H_dot plot.

    Returns
    -------
    dict
        Result dictionary with keys:
        - "COP" : float
            Coefficient of performance of the heat pump.
        - "output" : dict
            Component outputs and the configuration used ("config").
        - "warnings" : dict
            Warnings per component (and additional items, e.g., pressure_ratio).
        - "figure" : matplotlib.figure.Figure or None
            Figure object, if plotting=True.
        - "axes" : matplotlib.axes.Axes or None
            Axes object, if plotting=True.
        - "costs" : any
            Aggregate cost metric from components (from cb.orc_comp.all_costs).

    Raises
    ------
    KeyError
        If the compressor power is missing in the configuration
        (process.fixed.compressor.power).
    """
    new_config = kwargs.get("config", None)
    verbose = kwargs.get("verbose", False)
    plotting = kwargs.get("plotting", False)

    fig = None
    ax = None
    warnings = {}
    outputs = {}

    def add_w_o(comp):
        # collect warnings/outputs per component
        warnings[comp.name] = copy.deepcopy(comp.warning)
        outputs[comp.name] = copy.deepcopy(comp.output)

    # Build isolated, merged configuration
    config = resolve_config(dir_name, new_config)

    # Preliminary start state (mass flow will be set after the compressor)
    start0 = cb.comp.Start("start", config, m_dot=10e-3, verbose=verbose)

    # Compressor power from configuration
    try:
        power = config["process"]["fixed"]["compressor"]["power"]
    except Exception:
        raise KeyError("Configuration is missing process.fixed.compressor.power.")

    # Compressor
    compressor = cb.comp.FlowMachine("compressor", config, verbose=verbose)
    p_high = config["working_fluid"]["p_high"]
    p_ratio = p_high / start0.output["state_out"]["working_fluid"][1]
    if p_ratio < 1.0:
        warnings["pressure_ratio"] = SimpleNamespace(
            value=10.0 / p_ratio,
            message="Pressure ratio is wrong"
        )

    compressor.calculate(
        start0.output["state_out"],
        {"working_fluid": [600, p_high, 5e5]},
        run_param={"power": power}
    )

    # Working-fluid mass flow
    m_dot_w = compressor.output["m_dot"]["working_fluid"]
    m_dot_param = {"m_dot": {"working_fluid": m_dot_w}}

    # Start state with correct mass flow
    start = cb.comp.Start("start", config, m_dot=m_dot_w, verbose=verbose)
    add_w_o(start)
    add_w_o(compressor)

    # Condenser
    condenser = cb.comp.StaticHeatExchanger("condenser", config, verbose=verbose)
    inp_cond, _ = condenser.set_in_out({"working_fluid": compressor.output["state_out"]["working_fluid"]})
    condenser.calculate(in_states=inp_cond, run_param={"m_dot": {"working_fluid": m_dot_w}}, verbose=False)
    volumes_c = condenser.calculate_volume()
    if verbose:
        print(f"Storage Volumes condenser: {volumes_c}")
    add_w_o(condenser)

    # Throttle
    throttle = cb.comp.Throttle("throttle", config, verbose=verbose)
    throttle.calculate(condenser.output["state_out"], compressor.output["state_in"], run_param=m_dot_param)
    add_w_o(throttle)

    # Evaporator
    evaporator = cb.comp.StaticHeatExchanger("evaporator", config, verbose=verbose)
    inp_ev_wf, _ = evaporator.set_in_out({"working_fluid": throttle.output["state_out"]["working_fluid"]})
    inp_ev_sec, out_ev_sec = evaporator.set_in_out(start.output["state_in"], False)
    evaporator.calculate(inp_ev_wf, out_ev_sec, run_param={"m_dot": {"working_fluid": m_dot_w}}, verbose=False)
    volumes_e = evaporator.calculate_volume()
    if verbose:
        print(f"Storage Volumes evaporator: {volumes_e}")
    add_w_o(evaporator)

    # Performance figure
    cop = np.abs(condenser.output["q_dot"] / power)
    if verbose:
        print(f"COP: {cop:.4f}")

    costs = cb.orc_comp.all_costs([compressor, condenser, throttle, evaporator],
                                  verbose)

    # Combined plot (ensure all elements share the same figure/axes)
    if plotting:
        fig, ax = plt.subplots(1)
        # Copy plot defaults (plain values), then attach fig/ax
        plot_info = copy.deepcopy(cb.CB_DEFAULTS["Components"]["Plot"])
        plot_info["ax"] = ax
        plot_info["fig"] = fig
        plot_info["x-shift"] = [0, 0]

        # Start point: use shallow copy to preserve fig/ax references
        start_plot_info = plot_info.copy()
        start_plot_info.update({
            "label": ["start", ""],
            "col": ["ok", "bv"],
            "direction": 1
        })
        shift, direct = start.plot(start_plot_info)

        # Subsequent components on the same axes
        plot_info["x-shift"] = shift
        plot_info["direction"] = direct

        plot_info.update({
            "label": [compressor.name, ""],
            "col": ["-r", "bv-"]
        })
        shift, direct = compressor.plot(plot_info)

        plot_info["x-shift"] = shift
        plot_info["direction"] = direct
        plot_info.update({
            "label": [condenser.name, ""],
            "col": [":r", "rv-"]
        })
        shift, direct = condenser.plot(plot_info)

        plot_info["x-shift"] = shift
        plot_info["direction"] = direct
        plot_info.update({
            "label": [throttle.name, ""],
            "col": ["-b", "bv-"]
        })
        shift, direct = throttle.plot(plot_info)

        plot_info["x-shift"] = shift
        plot_info["direction"] = direct
        plot_info.update({
            "label": [evaporator.name, ""],
            "col": [":b", "bv-"]
        })
        evaporator.plot(plot_info)

    # Collect results
    outputs["config"] = copy.deepcopy(config)
    results = {
        "COP": cop,
        "output": outputs,
        "warnings": warnings,
        "figure": fig,
        "axes": ax,
        "costs": costs
    }
    return results

if __name__ == "__main__":
    dir_name_out = cb.CB_DEFAULTS["General"]["CB_DATA"] + "\\io-hp-data.yaml"

    conf_m = {
        "cold_storage": {"temp_low": 250.0},
        "working_fluid": {"p_high": 35e5}
        }

    res_act = heat_pump(dir_name_out, verbose=True, plotting=True)
    if any(ns.value != 0 for ns in res_act["warnings"].values()):
        print(f"Check warnings, at least one deviates from 0!\n{res_act['warnings']}")



# -*- coding: utf-8 -*-
"""
Calculation of a Carnot Battery (PTES) with two energy storages.


Created on Tue Aug  5 12:33:52 2025

@author: atakan
Universität Duisburg-Essen, Germany

In the framework of the Priority Programme: "Carnot Batteries: Inverse Design from
Markets to Molecules" (SPP 2403)
https://www.uni-due.de/spp2403/
https://git.uni-due.de/spp-2403/residuals_weather_storage

"""
import copy
import carbatpy as cb


def cb_calc(dir_names,  **kwargs):
    """
    Runs a Carnot battery model by sequentially simulating a heat pump and an Organic Rankine Cycle (ORC)
    using the provided configuration files or dictionaries.

    The function first executes the heat pump cycle based on the configuration in `dir_names["hp"]`.
    If successful and without warnings, it extracts the relevant thermodynamic conditions and storage states
    to configure and launch the subsequent ORC calculation (using `dir_names["orc"]`).
    Both cycles may accept additional configuration dictionaries (useful for parameter optimization).
    If `plotting` is True, figures for both cycles are generated.

    The overall Carnot battery round-trip efficiency (`rte`) is computed as the product of the heat pump COP
    and the ORC thermal efficiency.

    Parameters
    ----------
    dir_names : dict
        A dictionary with keys "hp" (for heat pump) and "orc" (for ORC), each containing
        either the path to a YAML (or similar) configuration file or a configuration dictionary.
    **kwargs :
        Additional keyword arguments. Supported keys:
            config : dict, optional
                Dictionary with additional or overriding configuration for the heat pump (`"hp"`) and/or ORC (`"orc"`)
                in the form {"hp": config_hp_dict, "orc": config_orc_dict}. Useful for optimization or parameter studies.
            verbose : bool, optional
                If True, enables verbose output. Default is False.
            plotting : bool, optional
                If True, generates diagrams for both the heat pump and ORC cycles.

    Returns
    -------
    rte : float
        The round-trip efficiency of the Carnot battery (heat pump COP × ORC thermal efficiency);
        returns negative values if warnings occur.
    results : dict
        A dictionary with two entries:
            "hp"  : results {COP, outputs_hp, warnings_hp, fig_hp, ax_hp}
            "orc" : results {eta_th, outputs_orc, warnings_orc, fig_orc, ax_orc}
        Each tuple contains the cycle performance metric, outputs, component warnings, and (optionally) plots.

    Notes
    -----
    This function calls both the heat pump and ORC functions in sequence, using outputs from the heat pump
    (such as working fluid composition and storage temperatures) to configure the ORC run. It is suitable 
    for integrating the full Carnot battery workflow, either for direct calculations, sensitivity studies, 
    or optimization routines.

    If component warnings deviate from zero for either cycle, the function will print diagnostic messages
    (if `verbose` is True) and return a negative code for `rte`.

    Example
    -------
    >>> dir_names = {"hp": "heat_pump_config.yaml", "orc": "orc_config.yaml"}
    >>> rte, results = cb_calc(dir_names, config={"hp": hp_opt_config, "orc": orc_opt_config}, plotting=True)
    """
    new_config = kwargs.get(
        "config", {"hp": None, "orc": None})  # for optimizations
    verbose = kwargs.get("verbose", False)
    plotting = kwargs.get("plotting", False)
    results = {}
    hp_src = dir_names["hp"] 
    if isinstance(hp_src, dict): 
        hp_src = copy.deepcopy(hp_src)
    orc_src = dir_names["orc"]
    if isinstance(orc_src, dict):
        orc_src = copy.deepcopy(orc_src)

    results["hp"] = cb.hp_comp.heat_pump(
        hp_src,
        config=copy.deepcopy(new_config["hp"]),
        verbose=verbose,
        plotting=plotting)

    if any(ns.value != 0 for ns in results["hp"]["warnings" ].values()):
        if verbose:
            print(f"Check HP Warnings, at least one deviates from 0!\n {results['hp']['warnings' ]}")
        return -sum(item.value for item in results["hp"]["warnings" ].values()), results

    # print(f"COP: {cop_h:.3f}")
    q_h = -results["hp"]["output" ]["condenser"]["q_dot"]
    
    # same fluid, same storage temperatures:
    orc_config = copy.deepcopy(cb.utils.io_utils.read_config(orc_src))
    from_hp_config = {"working_fluid": {"fractions": copy.deepcopy(results["hp"]["output" ]["config"]["working_fluid"]["fractions"]),
                                        "species": copy.deepcopy(results["hp"]["output" ]["config"]["working_fluid"]["species"])},
                      "cold_storage":  {"temp_low": copy.deepcopy(results["hp"]["output" ]["config"]["cold_storage"]["temp_low"])},
                      "hot_storage": {"temp_high": copy.deepcopy(results["hp"]["output" ]["config"]["hot_storage"]["temp_high"])}}
    merged_orc = copy.deepcopy(orc_config)
    for key in from_hp_config:
        merged_orc.setdefault(key, {})
        merged_orc[key].update(copy.deepcopy(from_hp_config[key]))

    results["orc"] = cb.orc_comp.orc(merged_orc,
                                     results["hp"]["COP" ],
                                     q_h,
                                     config=copy.deepcopy(new_config["orc"]),
                                     plotting=plotting,
                                     verbose=verbose)
    
    if any(ns.value != 0 for ns in results["orc"]['warnings'].values()):
        if verbose:
            print(f'Check ORC Warnings, at least one deviates from 0!\n {results["orc"]["warnings"]}')
        return -sum(item.value for item in results["orc"]['warnings'].values()), results
    rte = results["orc"]['eta_th']*results['hp']['COP']
    print(f"COP: {results['hp']['COP' ]:.3f}, eta: {results['orc']['eta_th']:.3f}, rte: {rte:.3f}")
    return rte, results


if __name__ == "__main__":
    dir_names_both = {"hp": cb.CB_DEFAULTS["General"]["CB_DATA"]+"\\io-hp-data.yaml",
                      "orc": cb.CB_DEFAULTS["General"]["CB_DATA"]+"\\io-orc-data.yaml"}
    rte, results_o = cb_calc(dir_names_both, plotting=True, verbose=True)
    print(f"RTE: {rte:.3f}\nHP, ORC",
          results_o["hp"]['warnings'],
          results_o["orc"]['warnings'])



# %% [markdown]
# A heat exchanger example
# ========================
# within carbatpy

# %%
import numpy as np
import matplotlib.pyplot as plt
import carbatpy as cb

# %% [markdown]
# Select two fluid (mixtures) and further conditions. H_DOT is the wanted heatflow rate ($\dot Q = $1 kW) here.

# %%
FLUID = "Propane * Pentane"  # working fluid
FLS = "Methanol"  # "Water"  # secondary fluid
comp = [.50, 0.5]
flm = cb.fprop.FluidModel(FLUID)
myFluid = cb.fprop.Fluid(flm, comp)

secFlm = cb.fprop.FluidModel(FLS)
secFluid = cb.fprop.Fluid(secFlm, [1.])
D_TEMP_MIN = 5.0

# Condenser, working fluid fixes all, secondary output enthalpy can be varied:
SEC_TEMP_IN = 300.0  # K
SEC_TEMP_OUT_MAX = 370.0
SEC_PRES_IN = 5e5  # Pa
H_DOT = 1e3  # W

# %% [markdown]
# Now set the wanted states, determine the enthalpies (alwas the value state[2] below), needed as input

# %%
state_sec_out = secFluid.set_state([SEC_TEMP_OUT_MAX, SEC_PRES_IN], "TP")

state_sec_in = secFluid.set_state(
    [SEC_TEMP_IN, SEC_PRES_IN], "TP")  # this is the entering state

# working fluid

TEMP_SAT_VAP = SEC_TEMP_OUT_MAX + D_TEMP_MIN
state_in = myFluid.set_state(
    [TEMP_SAT_VAP, 1.], "TQ")  # find minimum pressure

WF_TEMP_IN = TEMP_SAT_VAP + D_TEMP_MIN
WF_TEMP_OUT = SEC_TEMP_IN + D_TEMP_MIN
state_out = myFluid.set_state([WF_TEMP_OUT, state_in[1]], "TP")

state_in = myFluid.set_state([myFluid.properties.pressure,
                              WF_TEMP_IN],
                             "PT")

# %% [markdown]
# Now we create an instance of the heat exchanger here: hex0. The input are the two Fluid-instances with the correct input states. The enthalpies of the wanted output states (one will be varied later to meet the minimum approach temperature). The pressures will be deduced from the actual values.

# %%
hex0 = cb.hex_th.StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
                           state_sec_out[2],
                           d_temp_separation_min=D_TEMP_MIN)

# %% [markdown]
# It is called find_pinch, but it actually just varies the fluid mass flow rate ratio, until the minimum approach temperature(d_temp_separation_min) is reached, if it is possible for the given inlet, wanted outlet values. If not a warning is given.

# %%
hex0.find_pinch()
if hex0.warning > 0:
    print(hex0.warning_message)

# now plotting can directly be done in pinch_calc 2024-05-24
fig_act, ax_act = plt.subplots(1)
PLOT_INFO = {"fig": fig_act, "ax": ax_act, "what": [2, 0], "col": ["r:", "ko"],
             "label": ["work,c", "sec,c"], "x-shift": [0, 0]}
hex0.plot_info=PLOT_INFO
hex0.pinch_calc(verbose=True)
ax_act.legend()

# %%
print(hex0.warning, hex0.warning_message)

# %% [markdown]
# You can plot a second heat exchanger into the same plot by passing the figure information. But first we need a new instance of an heat exchanger, now an evaporator called hex1:

# %%
#  Evaporator: ----------------------------

SEC_TEMP_IN = 300.0 # K
SEC_TEMP_OUT = 285
SEC_PRES_IN = 15e5
H_DOT = 1e3  # heat flow rate
extra = 2  # some additional temperature difference between both fluids at the entrance of the working fluid
# D_TEMP_SUPER = 5.
D_TEMP_MIN = 6.0  # Minimum approach temperature, allowed

state_sec_out = secFluid.set_state([SEC_TEMP_OUT, SEC_PRES_IN], "TP")
state_out = myFluid.set_state([SEC_TEMP_IN-D_TEMP_MIN - extra, 1.0], "TQ")

# the inlet states must be the last set_state before the hex is constructed, because it is further used by the heat exchanger:
state_sec_in = secFluid.set_state([SEC_TEMP_IN, SEC_PRES_IN], "TP")
state_in = myFluid.set_state(
    [SEC_TEMP_OUT-D_TEMP_MIN - extra, state_out[1]], "TP")

print("state in (T,p,h,v,s,q,u):", state_in)

# %% [markdown]
# The new states are set, now we can do the calculations again.

# %%
hex1 = cb.hex_th.StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
                           state_sec_out[2],
                           d_temp_separation_min=D_TEMP_MIN)
# ms1, d_tempall1, w1, s1 = hex1.pinch_calc()

factor_out = hex1.find_pinch()
if hex1.warning > 2:
    print("Second heat exchanger:", hex1.warning_message, hex1.dt_min)
else:

    # plotting in the same figure
    PLOT_INFO = {"fig": fig_act, "ax": ax_act, "what": [2, 0], "col": ["k:", "bo"],
                 "label": ["work,e", "sec,e"], "x-shift": [0, 0]}
    hex1.plot_info=PLOT_INFO
    hex1.pinch_calc(verbose=True)
    ax_act.legend()

# %% [markdown]
# By the way, the mean temperature difference between both curves is evaluated. If you have an idea of the mean heat transfer coefficient U, you can calculate the needed heat exchanger area:

# %%
u_heat_transfer = 500 # W/ (m2 K)
area_needed = H_DOT/ np.abs(hex1.dt_mean) /u_heat_transfer
print(f"Estimated HEX area: {area_needed:.3f} m^2")

# %%
fig_act

# %%






# -*- coding: utf-8 -*-
"""
An example of using the classes in comp to calculate a heat pump.

Created on Mon Aug 12 17:28:32 2024

@author: atakan
"""

import matplotlib.pyplot as plt
import carbatpy as cb


if __name__ == "__main__":
    # in the following yaml file most parameters are set
    dir_name_out = cb.CB_DEFAULTS["General"]["CB_DATA"]+"\\io-cycle-data.yaml"
    
    #if you want to replace some of the values of the yaml file you can wrinte
    # them in a dictionary:
    conf_m = {"cold_storage": {"temp_low": 274.},
              "working_fluid": {"p_high": 1.49e6,
                                'fractions': [.4, 0.4, .20, 0.0000]},
              "process":{"fixed": {"compressor": {"power": 1000000.0}}}
              }
    
    # pass them for the calculation of the results
    res_act = cb.hp_comp.heat_pump(
        dir_name_out, config=conf_m, verbose=True, plotting =True)
    if any(ns.value != 0 for ns in res_act['warnings'].values()):
        print(f"Check Warnings, at least one deviates from 0!\n {res_act['warnings']}")
    # many results are now stored in the dictionary res_act, e.g.:
    print(f"\nThe COP is: {res_act['COP']:.2f}")





# -*- coding: utf-8 -*-
"""
Serialisation and storing dictionaries

Some German explanations from ChatGPT:

Um die Funktionalität für die Serialisierung und Deserialisierung sowie die Dateioperationen in mehreren Klassen zu verwenden, kannst du eine Basisklasse erstellen, die diese Methoden enthält. Andere Klassen können dann von dieser Basisklasse erben. Dadurch musst du den Code nicht in jede Klasse kopieren, sondern kannst die Methoden einfach durch Vererbung wiederverwenden.

Hier ist ein Beispiel, wie du dies machen kannst:
Basisklasse erstellen

Erklärung

    Basisklasse Serializable:
        Diese Klasse enthält alle Methoden und Eigenschaften, die für die Serialisierung, Deserialisierung und Dateioperationen erforderlich sind.
        Andere Klassen können einfach von dieser Basisklasse erben, um diese Funktionalitäten zu nutzen.

    Erben von Serializable:
        Die Klassen AnotherClass und YetAnotherClass erben von Serializable.
        Sie erhalten dadurch automatisch Zugriff auf alle Methoden und Eigenschaften von Serializable.

    Verwendung:
        Instanzen der abgeleiteten Klassen können erstellt, in JSON- oder YAML-Dateien gespeichert und wieder aus diesen Dateien geladen werden, ohne dass der Code für diese Funktionalitäten in jeder Klasse wiederholt werden muss.

Durch die Verwendung einer Basisklasse und Vererbung wird dein Code modularer, wiederverwendbarer und leichter zu warten.


Bei der Basisklasse steht explizit "AnotherClass", das muss doch ein Fehler sein, oder?
ChatGPT

Du hast recht, das war ein Fehler. Die Referenz auf AnotherClass in der Methode from_dict sollte generisch sein, damit sie für jede abgeleitete Klasse funktioniert. Hier ist die korrigierte Version:
Erklärung der Änderungen

    Generische Deserialisierung:
        In der Methode from_dict wird die Deserialisierung nun allgemein gehalten. Wenn ein Dictionary ein 'value'-Schlüssel hat, wird angenommen, dass es sich um eine spezielle Struktur handelt, die ebenfalls deserialisiert werden muss. Hier kann die Logik angepasst werden, um spezifische Klassen zu unterstützen.

    Serializable.from_dict:
        Um eine rekursive Deserialisierung zu ermöglichen, wurde Serializable.from_dict verwendet. Du kannst dies anpassen, um spezifische Klassen zu unterstützen, indem du den Typ des Dictionaries überprüfst.

Mit diesen Anpassungen ist die Basisklasse Serializable generisch genug, um in verschiedenen Klassen verwendet zu werden, ohne dass spezifische Anpassungen in der Basisklasse erforderlich sind. Die Logik zur Deserialisierung kann in den abgeleiteten Klassen spezifiziert werden, falls erforderlich.

Created on Fri Jul 26 11:22:13 2024

@author: atakan
"""
import json
import numpy as np
import yaml
import os

from typing import NamedTuple

class DataAll:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
            
    def print_attributes(self):
        for key, value in vars(self).items():
            print(f"{key}: {value}")
            
    def to_dict(self):
        return vars(self)


class DataNode:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if isinstance(value, dict):
                # Wenn der Wert ein Dictionary ist, rekursiv eine DataNode erstellen
                setattr(self, key, DataNode(**value))
            else:
                setattr(self, key, value)
    
    def to_dict(self):
        result = {}
        for key, value in vars(self).items():
            if isinstance(value, DataNode):
                result[key] = value.to_dict()
            else:
                result[key] = value
        return result

    def __getattr__(self, name):
        # Standardverhalten, falls ein Attribut nicht existiert
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

class Serializable:
    default_path = ''
    default_filename = 'default.json'
    default_storage_filename = 'def_store.json'

    def _is_serializable(self, value):
        try:
            json.dumps(value)
            return True
        except (TypeError, OverflowError):
            return False

    def _to_serializable(self, value):
        if isinstance(value, np.ndarray):
            return value.tolist()
        elif isinstance(value, (np.float64, np.int64)):
            return value.item()
        elif isinstance(value, dict):
            return {k: self._to_serializable(v) for k, v in value.items()}
        elif isinstance(value, list):
            return [self._to_serializable(item) for item in value]
        elif isinstance(value, (int, float, str, bool)):
            return value
        elif hasattr(value, 'to_dict'):
            return value.to_dict()
        else:
            return str(value)

    def _filter_components(self, components):
        def filter_dict(d):
            if isinstance(d, dict):
                return {k: filter_dict(v) for k, v in d.items() if k not in ('fluid', 'fluids', 'all_states')}
            return d

        return {k: filter_dict(v) for k, v in components.items()}

    def to_dict(self):
        result = {}
        for key, value in self.__dict__.items():
            if key not in ('all', 'fluid', 'fluids', 'all_states'):
                if key == 'components':
                    filtered_components = self._filter_components(value)
                    result[key] = {k: self._to_serializable(v) for k, v in filtered_components.items()}
                else:
                    result[key] = self._to_serializable(value)
        return result

    @classmethod
    def from_dict(cls, data):
        instance = cls.__new__(cls)
        for key, value in data.items():
            if isinstance(value, list) and len(value) > 0 and isinstance(value[0], (int, float)):
                setattr(instance, key, np.array(value))
            elif isinstance(value, dict):
                if key == 'components':
                    setattr(instance, key, cls._deserialize_components(value))
                else:
                    setattr(instance, key, value)
            else:
                setattr(instance, key, value)
        return instance

    @classmethod
    def _deserialize_components(cls, components):
        def deserialize_dict(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = deserialize_dict(v)
                elif isinstance(v, list):
                    result[k] = np.array(v) if all(isinstance(i, (int, float)) for i in v) else v
                else:
                    result[k] = v
            return result

        return {k: deserialize_dict(v) for k, v in components.items()}

    def save_to_file(self, file_path=None):
        if file_path is None:
            file_path = os.path.join(self.default_path, self.default_storage_filename)
        try:
            data = self.to_dict()
            if file_path.endswith('.json'):
                with open(file_path, 'w') as file:
                    json.dump(data, file, indent=4)
            elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
                with open(file_path, 'w') as file:
                    yaml.dump(data, file, default_flow_style=False)
            else:
                raise ValueError("Unsupported file format. Use .json or .yaml/.yml")
        except Exception as e:
            print(f"Error saving to file: {e}")

    @classmethod
    def load_from_file(cls, file_path=None):
        if file_path is None:
            file_path = os.path.join(cls.default_path, cls.default_filename)
        try:
            if file_path.endswith('.json'):
                with open(file_path, 'r') as file:
                    data = json.load(file)
            elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
                with open(file_path, 'r') as file:
                    data = yaml.load(file, Loader=yaml.SafeLoader)
            else:
                raise ValueError("Unsupported file format. Use .json or .yaml/.yml")

            return cls.from_dict(data)
        except Exception as e:
            print(f"Error loading from file: {e}")
            return None


# # Beispielklasse HpVal
# class HpVal(Serializable):
#     default_path = 'default_directory'
#     default_filename = 'default_file.json'
#     default_storage_filename = 'default_file_last.json'

#     def __init__(self, variables_dict=None):
#         if variables_dict:
#             for key, value in variables_dict.items():
#                 setattr(self, key, value)
# import carbatpy as cb
# #from carbatpy.heat_pump import set_fl_state
# # Beispielklasse HeatPump
# class HeatPump(Serializable):
#     default_path = 'default_directory'
#     default_filename = 'default_file_hp.json'
#     default_storage_filename = 'default_file_hp_last.json'

#     def __init__(self, fixed_points, components=None):
#         if components:
#             self.components = components
#         else: 
#             self.components = {}
#         self.fixed_points = fixed_points
#         self.fluids_ini = fixed_points["fluids_all"]
#         self.fluids = [cb.hp_simple.set_fl_state([fll, *ii], fixed_points)[0] for fll, ii in self.fluids_ini.items()]
#         self.all_states = []
#         self.m_dots = []
#         self.warning = []
#         self.evaluation = {
#             "Q_DOT_MIN": self.fixed_points["Q_DOT_MIN"],
#             "Power": 0.0,
#             "T_OUT_STORAGE_HOT": self.fixed_points["T_OUT_STORAGE_HOT"],
#             "T_OUT_STORAGE_COLD": self.fixed_points["T_OUT_STORAGE_COLD"],
#             "exergy_loss_rate": 0
#         }

if __name__ == "__main__":
    # DataAll usage;
    warning =DataAll(value=0, message= "All o.k.")
    warning.print_attributes()
    
    
    # Testen
    fixed_points = {
        "fluids_all": {"fluid1": ["Propane", 0.675], "fluid2": ["Butane", 0.0515]},
        "Q_DOT_MIN": 1000.0,
        "T_OUT_STORAGE_HOT": 363.0,
        "T_OUT_STORAGE_COLD": 250.15
    }
    
    heat_pump = HeatPump(fixed_points)
    print("HeatPump attributes:")
    print(heat_pump.all)
    
    # Serialisieren und Speichern
    heat_pump.save_to_file('heat_pump.json')
    
    # Laden und Deserialisieren
    loaded_heat_pump = HeatPump.load_from_file('heat_pump.json')
    print("Loaded HeatPump attributes:")
    print(loaded_heat_pump.all)
    
    
    
    # Example:
    class AnotherClass(Serializable):
        default_path = '/path/to/another_class'
        default_filename = 'another_class.json'
    
        def __init__(self, attribute1, attribute2):
            self.attribute1 = attribute1
            self.attribute2 = attribute2
    
    class YetAnotherClass(Serializable):
        default_path = '/path/to/yet_another_class'
        default_filename = 'yet_another_class.json'
    
        def __init__(self, attribute3, attribute4):
            self.attribute3 = attribute3
            self.attribute4 = attribute4

if __name__ == "__main__":
    # Erstellen von Instanzen und Speichern in Dateien
    instance1 = AnotherClass(attribute1='value1', attribute2=42)
    instance2 = YetAnotherClass(attribute3=np.array([1, 2, 3]), attribute4=True)
    
    # Speichern mit Standardpfad und -dateinamen
    instance1.save_to_file()
    instance2.save_to_file()
    
    # Speichern mit benutzerdefiniertem Pfad und Dateinamen
    instance1.save_to_file('/custom/path/custom_filename.json')
    instance2.save_to_file('/custom/path/custom_filename.yaml')
    
    # Laden mit Standardpfad und -dateinamen
    loaded_instance1 = AnotherClass.load_from_file()
    loaded_instance2 = YetAnotherClass.load_from_file()
    
    # Laden mit benutzerdefiniertem Pfad und Dateinamen
    loaded_instance1 = AnotherClass.load_from_file('/custom/path/custom_filename.json')
    loaded_instance2 = YetAnotherClass.load_from_file('/custom/path/custom_filename.yaml')
    
    print(loaded_instance1.to_dict())
    print(loaded_instance2.to_dict())




# -*- coding: utf-8 -*-
"""
cycle calculation one input for each cyle
Created on Wed Jul 31 17:01:59 2024

@author: atakan
"""
import copy
import numpy as np
import matplotlib.pyplot as plt
import carbatpy as cb
from carbatpy.helpers.ser_dict import Serializable


class Cycle:
    def __init__(self, components=None, fixed_points=None,  **kwargs):
        self.components = components if components else {}
        self.fixed_points = fixed_points
        
        self.all_states = []
        self.m_dots = []
        self.evaluation = {}
        self.warning = []
        self.kwargs = kwargs
        
        self.fluids_all = copy.copy(fixed_points["fluids_all"])
        self.fluids = []
        self.set_fl_state()

    def calculate_hex(self, params, plot_info={}, verbose_=True):
        """
        General method for calculating a heat exchanger.

        params: dict containing the following keys:
            - fluid_numbers: List[int]
            - h_dot_min: float
            - h_out_w: float
            - h_limit_s: float (optional)
            - d_temp_separation_min: float (optional)
            - calc_type: str (optional)
            - pinch: str (optional)
            - name: str (optional)
        """
        fluid_numbers = params["fluid_numbers"]
        h_dot_min = params["h_dot_min"]
        h_out_w = params["h_out_w"]
        h_limit_s = params.get("h_limit_s", np.nan)
        d_temp_separation_min = params.get("d_temp_separation_min", 0.5)
        calc_type = params.get("calc_type", "const")
        pinch = params.get("pinch", "")
        name = params.get("name", "evaporator")

        w_actual, s_actual = [self.fluids[fluid_numbers[0]],
                              self.fluids[fluid_numbers[1]]]
        hex_act = cb.hex_th.StaticHeatExchanger(
            [w_actual, s_actual],
            h_dot_min, h_out_w, h_limit_s,
            d_temp_separation_min=d_temp_separation_min,
            name=name
        )

        if pinch == "":
            h_s_out = hex_act.pinch_calc(plot_info)
        elif pinch == "secondary":
            h_s_out = hex_act.find_pinch()
        elif pinch == "working":
            h_s_out = hex_act.find_pinch(False)
        else:
            print("calculate_hex: This option is not available")

        if hex_act.warning > 0:
            print("Heat-Exchanger problem: ", hex_act.warning, hex_act.warning_message)
            self.warning.append([hex_act.name, hex_act.warning, hex_act.warning_message])
        
        m_dot_s, d_tempall, wf_states, sf_states = hex_act.pinch_plot(plotting=verbose_)
        d_temp_mean = hex_act.dt_mean
        d_temp_min = hex_act.dt_min

        self.all_states.append(wf_states)
        self.all_states.append(sf_states)
        self.m_dots.append(hex_act.m_dot_w)
        self.m_dots.append(m_dot_s)

        w_after_hex = self.fluids[0].set_state([wf_states[0, 1], wf_states[0, 2]], "PH")
        fluids_states_in_out = np.array([[wf_states[0], wf_states[-1]], [sf_states[0], sf_states[-1]]])
        fluids_states_in_out_ex = fluids_states_in_out
        if hex_act.heating < 1:
            fluids_states_in_out_ex = np.flip(fluids_states_in_out, axis=1)

        mass_flow_rates = np.array([hex_act.m_dot_w, hex_act.m_dot_s])
        ex_loss_rate = cb.exlo.exergy_loss_flow(fluids_states_in_out_ex, mass_flow_rates)
        
        self.evaluation[name] = {
            "states-in-out": fluids_states_in_out,
            "mass-flow-rates": mass_flow_rates,
            "exergy-loss-rate": ex_loss_rate,
            "dT-min": d_temp_min,
            "dT-mean": d_temp_mean
        }
        self.components[name] = hex_act.all
        
        if verbose_:
            print(f"After heat exchanger: {name}")
            w_actual.print_state()
        
        return w_after_hex

    def calc_cycle(self):
        raise NotImplementedError("This method must be overridden in subclasses.")
        
    def plot(self):
        raise NotImplementedError("This method must be overridden in subclasses.")
        
class HeatPump(Cycle):
    def __init__(self, components=None, fixed_points=None, **kwargs):
        super().__init__(components=components, fixed_points=fixed_points, kwargs=kwargs)
        # Additional initializations specific to HeatPump
        
        
    def set_fl_state(self,
                      **kwargs):
        
        fluids =  {}
        p_all= {}
        for key in self.fluids_all.keys():
            fluids[key] = cb.init_fluid(self.fixed_points[key]["fluid"],
                                      self.fixed_points[key]["composition"],
                                      kwargs=self.fixed_points[key]) # props etc. cn be passed here
            
            p_all[key]= self.fixed_points[key]["p_low"]
            
        for compound in self.fixed_points["process"].keys():
            if compound =="compressor":
                dt_sup = self.fixed_points[compound]["dt_superheat"]
            elif compound =="throttle":
                pass
            elif compound == "condenser":
                t_in_out = self._find_st_in_out(compound)
                dt = self.fixed_points[compound]["dt_min"]
                for fl_act in self.fixed_points[compound]["fluid"].keys():
                    if "storage" in fl_act:
                        state_out = fluids[fl_act].set_state([t_in_out['out'], p_all[fl_act]], "TP")
                        state_in = fluids[fl_act].set_state([t_in_out['in'], p_all[fl_act]], "TP")
                        self.fixed_points['h_h_out_sec'] = state_out[2]
                    else:
                        state_out = fluids[fl_act].set_state([t_in_out['out']+dt, p_all[fl_act]], "TP")
                        
                
                            
            
        #     if "storage" in key:
        #         p_in = self.fixed_points[key]["p_low"]
        #         t_in_out =t_val["hot_storage"] # Default T_sur is used!
        #         state_out = fluid_act.set_state([t_in_out['out'], p_in], "TP")
        #         state_in = fluid_act.set_state([t_in_out['in'], p_in], "TP")
    
        #         if "hot" in key:
        #             self.fixed_points['h_h_out_sec'] = state_out[2]
                    
        #         elif "cold" in key:
        #             self.fixed_points['h_l_out_cold'] = state_out[2]
    
                    
    
        #     w_p_select = self.fixed_points["working_fluid"]
        #     if w_p_select["setting"]== "auto":
        #         temp_high= self.fixed_points['hot_storage']["temp_high"]
        #         state_in_cond = fluid_act.set_state(
        #             [temp_high, 1.], "TQ")  # find high pressure
        #         p_high = state_in_cond[1]
    
        #         state_out_cond = fluid_act.set_state([temp_high +
        #                                             self.fixed_points["condenser"]['dt_min'],
        #                                             p_high], "TP")
                
        #         T_IN = t_val["evaporator"] - fixed_points['D_T_MIN']
        #         state_satv_evap = fluid_act.set_state(
        #             [T_IN -
        #              fixed_points['D_T_SUPER'], 1.], "TQ")  # find minimum pressure
        #         p_low = state_satv_evap[1]
    
        #         state_out_evap = fluid_act.set_state([p_low,
        #                                             T_IN], "PT")
        #     else:
        #         p_high =  w_p_select["p_high"]
        #         p_low =  w_p_select["p_low"]
        #         state_in_cond = fluid_act.set_state(
        #             [fixed_points['T_OUT_STORAGE_HOT'],p_high], "TP")  # find high pressure
    
    
        #         state_out_cond = fluid_act.set_state([fixed_points['T_IN_STORAGE_HOT'] +
        #                                             fixed_points['D_T_MIN'],
        #                                             p_high], "TP")
        #         T_IN = fixed_points['T_IN_STORAGE_HOT'] - fixed_points['D_T_MIN']
        #         state_satv_evap = fluid_act.set_state(
        #             [T_IN -
        #              fixed_points['D_T_SUPER'], p_low], "TP")  # find minimum pressure
    
        #         state_out_evap = fluid_act.set_state([p_low,
        #                                             T_IN], "PT")
    
        #     fixed_points["p_low"] = p_low
        #     fixed_points["p_high"] = p_high
        #     fixed_points['h_h_out_w']= state_out_cond[2]
        #     fixed_points['h_l_out_w']= state_out_evap[2]
        #     self.fluids.append(fluid_act)
        # return fluids_act, fixed_points
    
    
    
    
    def _find_st_in_out(self, what,  t_amb =cb.CB_DEFAULTS['General']["T_SUR"]):
        
        for key in self.fixed_points[what]["fluid"]:
            if "storage" in key:
                fl_name = self.fixed_points[what]["fluid"]
                
    
            # for key in self.fixed_points["process"].keys():
            #     di2 = self.fixed_points[key]
                if fl_name in di2["fluid"].keys():
                    in_out = di2["fluid"][fl_name]
                    t_both= {}
                    for key, value in in_out.items():
                        t_act=  self.fixed_points[fl_name][value]
                        if t_act =="ambient":
                            t_act = t_amb
                        t_both[key]=t_act
                
                return t_both
    def _get_all_t(self):
        t_values={}
        for key in self.fixed_points:
            if "storage" in key:
                t_b = self._find_st_in_out(key)
                t_values[key]=t_b
        t_values["d_super"]=self.fixed_points["compressor"]["dt_superheat"]
        t_values["dt_min_cond"]=self.fixed_points["coondenser"]["dt_min"]
        t_values["dt_min_evap"]=self.fixed_points["evaporator"]["dt_min"]
        return t_values
                    
    

    def calc_cycle(self, plot_info={}, variations=["", ""], verbose=False):
        """
        Calculates a simple compression heat pump cycle.
        """
        state_in = copy.copy(self.fluids[0])
        w_actual = self.fluids[0]
        if verbose:
            print("Start:")
            w_actual.print_state()

        # Compressor
        compressor = cb.FlowDevice(
            self.fluids[0], self.fixed_points["p_high"], 1.0,
            device_type="machine", name="compressor",
            calc_type="const_eta", calc_parameters={"eta_s": self.fixed_points["ETA_S_C"]},
            plot_info=plot_info
        )
        w_after_compressor, work_c, _power = compressor.state_w_p()
        self.components[compressor.name] = compressor.all
        
        if verbose:
            print("After the compressor")
            w_actual.print_state()

        # Condenser
        params = {
            "fluid_numbers": [0, 1],
            "h_dot_min": self.fixed_points["Q_DOT_MIN"],
            "h_out_w": self.fixed_points["h_h_out_w"],
            "h_limit_s": self.fixed_points["h_h_out_sec"],
            "d_temp_separation_min": self.fixed_points['D_T_MIN'],
            "calc_type": variations[0],
            "name": "condenser"
        }
        w_after_condenser = self.calculate_hex(params, plot_info=plot_info, verbose_=verbose)

        power = work_c * self.m_dots[params["fluid_numbers"][0]]
        heat_flow_evap = self.fixed_points["Q_DOT_MIN"] - power
        if verbose:
            print(f"Power: {power} W, Evaporator heat: {heat_flow_evap} W\n")

        # Throttle
        throttle = cb.FlowDevice(
            w_actual, self.fixed_points["p_low"], self.m_dots[params["fluid_numbers"][0]],
            device_type="throttle", name="throttle-A", calc_type="const_h", calc_parameters={},
            plot_info=plot_info
        )
        w_after_throttle = throttle.output["state_out"]
        self.components[throttle.name] = throttle.all

        if verbose:
            print("After the throttle:")
            print(w_actual.print_state())

        w_act = w_actual.set_state([w_after_throttle[1], w_after_throttle[2]], "PH")
        if self.fixed_points["T_OUT_STORAGE_COLD"] < w_act[0] + self.fixed_points['D_T_MIN']:
            self.fixed_points["T_OUT_STORAGE_COLD"] = w_act[0] + self.fixed_points['D_T_MIN']

        # Evaporator
        params = {
            "fluid_numbers": [0, 2],
            "h_dot_min": heat_flow_evap,
            "h_out_w": self.fixed_points["h_l_out_w"],
            "h_limit_s": self.fixed_points["h_l_out_cold"],
            "d_temp_separation_min": self.fixed_points['D_T_MIN'],
            "calc_type": variations[0],
            "name": "evaporator"
        }
        w_after_evap = self.calculate_hex(params, plot_info=plot_info, verbose_=verbose)

        q_evap = (w_after_evap[2] - w_after_throttle[2]) * self.m_dots[params["fluid_numbers"][0]]
        if heat_flow_evap - q_evap > 1:
            print(f"Heat pump cycle not in equilibrium: {heat_flow_evap}, {q_evap}")

        if verbose:
            print(f"Evaporator heat flow rate: {q_evap}, Power: {power}")
            w_actual.print_state()

        # Calculate exergy losses
        sec_fluids_states_in_out = np.array([
            self.evaluation["condenser"]["states-in-out"][-1],
            self.evaluation["evaporator"]["states-in-out"][-1]
        ])
        mass_flow_rates = np.array([
            self.evaluation["condenser"]["mass-flow-rates"][-1],
            self.evaluation["evaporator"]["mass-flow-rates"][-1]
        ])
        ex_loss_rate = cb.exlo.exergy_loss_flow(sec_fluids_states_in_out, mass_flow_rates)

        cop = self.fixed_points["Q_DOT_MIN"] / power
        self.evaluation.update({
            "Q_DOT_MIN": self.fixed_points["Q_DOT_MIN"],
            "Power": power,
            "p_high": self.fixed_points["p_high"],
            "p_low": self.fixed_points["p_low"],
            "exergy_loss_rate": ex_loss_rate,
            "sec_fluid_states_in_out": sec_fluids_states_in_out,
            "sec_fluid_m_dots": mass_flow_rates,
            "COP": cop
        })
        return cop
    
    def plot(self, f_name=None, **kwargs):
    
        """
        plots the heat pump cycle and stores it to the given file (name)
    
        Returns
        -------
        None.
    
        """
        if not f_name:
            f_name =cb.CB_DEFAULTS["General"]["RES_DIR"]+"\\last_T_H_dot_plot_hp"
        if kwargs:
            print(f"keyword argumnts not implemented in hp-plot, will be ignored{kwargs}")
        states = []
        connect_temp = []
        connect_h = []
        position = [0, -1]
        fig, ax = plt.subplots(1, 1)
        points = ["-k", "or", "-b", "vg"]
        m_dots = self.m_dots
        n_val = [2, 0]  # positions of h and T
        n_points = np.shape(self.all_states)[1]
        n_range = np.shape(m_dots)[0]  # how many streams? 2 x heat exchangers
        data = np.zeros((n_points, 4))
        names = ["HP-process", "{$H_{dot} /W", "T / K", "$m_{dot}"]
        all_data = []
    
        m_dots[3] = m_dots[3]  # / m_dots[2] * m_dots[0]
        m_dots[2] = m_dots[0]
    
        for i_process in range(4):
    
            cycle_states = self.all_states[i_process]
            cycle_states = cycle_states[cycle_states[:, 2].argsort(), :]
            if i_process == 2:
                liquid = self.fluids[0].set_state([cycle_states[0, 1], 0.],
                                                  "PQ")
                ax.plot((liquid[2] - cycle_states[0, 2]) * m_dots[i_process],
                        liquid[0], "o")
            cycle_states[:, 2] -= cycle_states[0, 2]
    
            cycle_states[:, 2] = cycle_states[:, 2] * m_dots[i_process]
            states.append(cycle_states)
    
            ax.plot(cycle_states[:, 2], cycle_states[:, 0], points[i_process])
            data[:, 0] = float(i_process)
            data[:, -1] = m_dots[i_process]
            data[:, 1:3] = cycle_states[:, n_val]
            all_data.append(copy.copy(data))
            if i_process == 0 or i_process == 2:
                connect_temp.append(cycle_states[position, 0])
                connect_h.append(cycle_states[position, 2])
        temperatures = np.array(connect_temp).reshape((4))
        enthalpies = np.array(connect_h).reshape((4))
        ax.set_xlabel('$ \\dot H$ / W')
        ax.set_ylabel("$T$ / K")
    
        for i_process in range(2):
            ax.plot([enthalpies[0+i_process], enthalpies[2+i_process]],
                    [temperatures[0+i_process], temperatures[2+i_process]], "r:")
        fig.savefig(f_name+".png")
        np.savetxt(f_name+".csv", np.concatenate(all_data),
                   delimiter=";", header=";".join(names))
        np.save(f_name+"_evaluation_dict.npy", self.evaluation)
        return [fig, ax]


class ORC(Cycle):
    def __init__(self, components=None, fixed_points=None, fluids=None):
        super().__init__(components, fixed_points, fluids)
        # Additional initializations specific to ORC

    def calc_cycle(self, verbose=False):
        """
        Calculates a simple Organic Rankine Cycle (ORC).
        """
        state_in = copy.copy(self.fluids[0])
        w_actual = self.fluids[0]
        if verbose:
            print("Start:")
            w_actual.print_state()

        # Pump
        pump = cb.FlowDevice(
            self.fluids[0], self.fixed_points["p_high"], 1.0,
            device_type="pump", name="pump",
            calc_type="const_eta", calc_parameters={"eta_s": self.fixed_points["eta_s_p"]}
        )
        w_after_pump, work_pump, _power = pump.state_w_p()
        m_dot_w = self.fixed_points["Q_DOT_MIN"] / (self.fixed_points["h_out_w"] - w_after_pump[2])
        self.components[pump.name] = pump.all

        if verbose:
            print("After the pump")
            w_actual.print_state()

        # Evaporator
        params = {
            "fluid_numbers": [0, 1],
            "h_dot_min": self.fixed_points["Q_DOT_MIN"],
            "h_out_w": self.fixed_points["h_out_w"],
            "d_temp_separation_min": self.fixed_points["D_T_MIN"],
            "calc_type": self.fixed_points["Q_DOT_MIN"],
            "name": "evaporator"
        }
        w_after_evaporator = self.calculate_hex(params, verbose_=verbose)

        # Expander
        expander = cb.FlowDevice(
            self.fluids[0], self.fixed_points["p_low"], m_dot_w,
            device_type="machine", name="expander",
            calc_type="const_eta", calc_parameters={"eta_s": self.fixed_points["eta_s_e"]}
        )
        w_after_expander, work_expander, _power = expander.state_w_p()
        self.components[expander.name] = expander.all

        if verbose:
            print("After the expander")
            w_actual.print_state()

        # Condenser
        params = {
            "fluid_numbers": [0, 2],
            "h_dot_min": self.fixed_points["Q_DOT_MIN"],
            "h_out_w": self.fixed_points["h_out_c"],
            "d_temp_separation_min": self.fixed_points["D_T_MIN"],
            "calc_type": self.fixed_points["Q_DOT_MIN"],
            "name": "condenser"
        }
        w_after_condenser = self.calculate_hex(params, verbose_=verbose)

        # Net work and efficiency
        work_net = work_expander - work_pump
        eta = work_net / self.fixed_points["Q_DOT_MIN"]

        self.evaluation.update({
            "NetWork": work_net,
            "p_high": self.fixed_points["p_high"],
            "p_low": self.fixed_points["p_low"],
            "Efficiency": eta
        })
        return eta


if __name__ =="__main__":
    import json
    file_n = cb.CB_DEFAULTS["General"]["RES_DIR"]+"\\test_input3.json"
    with open(file_n, "r") as file:
        inp_hp = json.load(file)
        
    file_n = cb.CB_DEFAULTS["General"]["CB_DATA"]+"\\orc_input.json"
    with open(file_n, "r") as file:
        inp_orc = json.load(file)
        
    hp_act =HeatPump(fixed_points=inp_hp)
    hp_act.calc_cycle()
    hp_act.plot()


# -*- coding: utf-8 -*-
"""
Created on Tue Jul 23 13:21:18 2024

@author: atakan
"""

import copy
from time import time
# import src.models.fluids.fluid_props as fprop
import numpy as np
from scipy.optimize import minimize, minimize_scalar, Bounds  # root, root_scalar
import matplotlib.pyplot as plt
import carbatpy as cb



def p_var_for_opt_hp(pressures,
                     fluids=None,
                     fixed=None,
                     t_val=None):
    
        state_out_cond = myFluid.set_state([t_val[0],
                                            pressures[1]], "TP")
        state_out_evap = myFluid.set_state([pressures[0],
                                            t_val[1]], "PT")
        fixed_new = {"eta_s": _ETA_S_,
                        "p_low": pressures[0],
                        "p_high": pressures[1],
                        "T_hh": t_val[2],
                        
                        "h_h_out_w": state_out_cond[2],
                        
                        "h_l_out_w": state_out_evap[2],
                        }
        fixed.update(fixed_new)
        # fixed["p_low"]= pressures[0]
        # fixed["p_high"]= pressures[1]
        hp_act = cb.hp_simple.HeatPump(fluids, fixed)
        print( "in opt:", pressures, t_val)  # hp_act.evaluation,
        cop = hp_act.calc_heat_pump(verbose=False)
        print(hp_act.warning, cop)
        if len (hp_act.warning) == 0: # for minimization
            return -cop
        return 10
    


if __name__ == "__main__":

    FLUID = "Propane * Butane * Pentane * Hexane"
    comp = [.75, 0.05, 0.15, 0.05]
    # comp = [0.4,	0.3,	0.3, 0.0]  # [0.164,.3330,.50300,0.0]

    FLS = "Water"  #
    FLCOLD = "Methanol"  # "Water"  #

    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])

    coldFlm = cb.fprop.FluidModel(FLCOLD)
    coldFluid = cb.fprop.Fluid(coldFlm, [1.])

    # Condenser(c) and storage (s), secondary fluids fix all, temperatures(T in K),
    # pressures (p in Pa)
    _ETA_S_ = 0.7  # interesting when changed from 0.69 to 0.65, the efficiency
    # decreases, the reason is the low quality along throtteling then
    _STORAGE_T_IN_ = cb._T_SURROUNDING
    _COLD_STORAGE_T_IN_ = _STORAGE_T_IN_
    _STORAGE_T_OUT_ = 363.  # 395.0
    _COLD_STORAGE_T_OUT_ = 250.15
    _STORAGE_P_IN_ = 5e5
    _COLD_STORAGE_P_IN_ = 5e5
    _Q_DOT_MIN_ = 1e3  # and heat_flow rate (W)
    _D_T_SUPER_ = 5  # super heating of working fluid
    _D_T_MIN_ = 4.  # minimum approach temperature (pinch point)
    # high T-storages
    state_sec_out = secFluid.set_state([_STORAGE_T_OUT_, _STORAGE_P_IN_], "TP")
    state_sec_in = secFluid.set_state([_STORAGE_T_IN_, _STORAGE_P_IN_], "TP")

    #  low T storages:
    state_cold_out = coldFluid.set_state(
        [_COLD_STORAGE_T_OUT_, _COLD_STORAGE_P_IN_], "TP")
    state_cold_in = coldFluid.set_state(
        [_COLD_STORAGE_T_IN_, _COLD_STORAGE_P_IN_], "TP")

    # working fluid
    T_DEW = _STORAGE_T_OUT_  # + _D_T_MIN_
    state_in_cond = myFluid.set_state([T_DEW, 1.], "TQ")  # find high pressure
    state_out_cond = myFluid.set_state([_STORAGE_T_IN_ + _D_T_MIN_,
                                        state_in_cond[1]], "TP")
    state_satv_evap = myFluid.set_state(
        [_STORAGE_T_IN_-_D_T_MIN_-_D_T_SUPER_, 1.], "TQ")  # find minimum pressure
    p_low = state_satv_evap[1]

    T_IN = _STORAGE_T_IN_ - _D_T_MIN_

    state_out_evap = myFluid.set_state([p_low,
                                        T_IN], "PT")

    FIXED_POINTS = {"eta_s": _ETA_S_,
                    "p_low": state_out_evap[1],
                    "p_high": state_in_cond[1],
                    "T_hh": _STORAGE_T_OUT_,
                    "h_h_out_sec": state_sec_out[2],
                    "h_h_out_w": state_out_cond[2],
                    "h_l_out_cold": state_cold_out[2],
                    "h_l_out_w": state_out_evap[2],
                    "T_hl": _STORAGE_T_IN_,
                    "T_lh": _STORAGE_T_IN_,
                    "T_ll": _COLD_STORAGE_T_OUT_,  # 256.0,
                    "Q_dot_h": _Q_DOT_MIN_,
                    "d_temp_min": _D_T_MIN_}

    print(
        f"p-ratio: {state_in_cond[1]/state_out_evap[1]: .2f}, p_low: {state_out_evap[1]/1e5: .2} bar")
    hp0 = cb.hp_simple.HeatPump([myFluid, secFluid, coldFluid], FIXED_POINTS)
    print(hp0.evaluation)
    cop = hp0.calc_heat_pump(verbose=True)
    print(hp0.evaluation)
    hp0.hp_plot()
    print(hp0.evaluation, "\n----------------\n")

    out = hp0.evaluation
    
    
    #----------
    # Test evaluation of a function to optimitze the two pressures of a heat pump:
    
    p_high = state_in_cond[1]
    pressures_act=[p_low, p_high]
    min_v = 0.95
    max_v =1.05 # bounds for optimization
    fluids=myFluid, secFluid, coldFluid
    fixed_act = FIXED_POINTS
    t_val=[_STORAGE_T_IN_ + _D_T_MIN_,
                T_IN,
                _STORAGE_T_OUT_]
    
    
    out_f= p_var_for_opt_hp(pressures_act,
                         fluids=fluids,
                         fixed = FIXED_POINTS,
                         t_val=t_val)
    print(f"opt-Function call. COP: {out_f}")
    
    bound_act = Bounds(lb =[p_low*min_v, p_high* min_v],
                       ub =[p_low*max_v, p_high* max_v])
    tolerance = 5e-3
    t0 = time()
    
    result = minimize(p_var_for_opt_hp,
                      pressures_act,
                      args=(fluids,
                            FIXED_POINTS,
                            t_val),
                      method='Nelder-Mead',
                      tol=tolerance,
                      bounds=bound_act,
                      options ={"maxiter": 24,
                                "disp": True})
    t1 =time()
    print(result, "time:",t1-t0)
    p_var_for_opt_hp(result.x,
                     fluids=fluids,
                     fixed = FIXED_POINTS,
                     t_val=t_val )
    
    
    #--------------------------
    
    
    print(f"Min and mean dT evaporator: {out['evaporator']['dT-min']}, {out['evaporator']['dT-mean']}")
    print(f"Min and mean dT condenser: {out['condenser']['dT-min']}, {out['condenser']['dT-mean']}")
    print(
        f"COP: {cop},p-ratio: {out['p_high']/out['p_low']:.2f}, p_low {out['p_low']/1e5:.2f} bar")

    print(
        f'exergy loss rate: {out["exergy_loss_rate"]}, eff: {1-out["exergy_loss_rate"]/out["Power"]:.4f}')

    #
    my_dict = cb.hp_simple.read_hp_results(cb._RESULTS_DIR+
                              "\\last_T_H_dot_plot_hp_evaluation_dict.npy")


# -*- coding: utf-8 -*-
"""
Example for Optimization of a Carnot battery, with a heat pump (heat_pump_comp) 
for charging and an ORC for discharging using scipy.optimize

Composition and high pressure (condenser) of the working fluid, and the low
temperature of the double tank cold storage are optimization variables for the heat pump. All
other parameters are fixed in a yaml file (io-cycle-data.yaml). For the ORC, the same
composition is used and only the two working fluid pressure levels are optimized.

Created on Mon Aug  4 13:12:24 2025

@author: atakan
Universität Duisburg-Essen, Germany

In the framework of the Priority Programme: "Carnot Batteries: Inverse Design from
Markets to Molecules" (SPP 2403)
https://www.uni-due.de/spp2403/
https://git.uni-due.de/spp-2403/residuals_weather_storage

"""

import numpy as np
import pandas as pd
import carbatpy as cb
import datetime
from pathlib import Path
import shutil
import yaml
import pickle


# Konfiguration und Konstanten (diese können außerhalb des main-Blocks stehen)

OPTI = True
HOW_OPT = "bas_hop"  #'shgo' # "dif_evol"  # "local" #
current_date = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-")
STORE_FILENAME = cb.CB_DEFAULTS["General"]["RES_DIR"] +"\\"+ current_date + "cb_opt_result"
POWER_C = 2000. # compressr power of heat pump
DIR = cb.CB_DEFAULTS["General"]["CB_DATA"]


# Ordner erstellen und YAML schreiben
dir_res = Path(STORE_FILENAME)
dir_res.mkdir(exist_ok=True)

datei = dir_res / 'config_bound.yaml'
res_dat = dir_res / 'opt_res.yaml'
file_res = dir_res / (current_date + "cb_opti_res.csv")


dir_names_both = {"hp": DIR+"\\io-hp-data.yaml",
                  "orc": DIR+"\\io-orc-data.yaml"}
for io_file in dir_names_both.values():
    shutil.copy2(io_file, dir_res)
    

# optimization variables, must be present in the file above
conf_hp = {"cold_storage": {"temp_low": 265.},
          "working_fluid": {"p_high": 1.45e6,
                            "p_low": 1.8e5,
                            'fractions': [.438, .4310, 0.131,  0.0000],
                            }
          }
conf_orc = {"working_fluid": {"p_high": 4.71e5,
                            "p_low": 2.21e5,
                            }
            }

# bounds of the optimization variables
bounds_hp = {"cold_storage": {"temp_low": [265, 277]},
            "working_fluid":
            {"p_high": [8.e5, 01.9e6],
             "p_low":[1e5, 2.9e5],
             'fractions': [[0.3, .85], [0.0, .7], [0.0, 0.3], [0, 0.25]]},
            }
bounds_orc = {"working_fluid":
            {"p_high": [4e5, 0.82e6],
             "p_low":[2.10e5, 3.9e5],},
            }

configs_m ={"hp" : conf_hp,
          "orc": conf_orc}
bounds_m = {"hp" : bounds_hp,
            "orc" : bounds_orc}

with datei.open('w') as f:
    yaml.dump({"configs":configs_m,
               "bounds":bounds_m,
               'power_compressor': POWER_C}, f, default_flow_style=False,
               sort_keys=False)

# WICHTIG: Der ausführbare Code muss in diesem Block stehen!
if __name__ == "__main__":
    if OPTI:
        # for optimization:
        print("Carnot battery optimization is running. It may take a while!")
        opt_res, paths = cb.opti_cycle_comp_helpers.optimize_cb(
            dir_names_both, POWER_C, configs_m, bounds_m,
            optimize_global=HOW_OPT,
            workers=1,
            maxiter=1,
            #opt_opt = -.03,
            )
        print(opt_res)
        with open(dir_res / "paths.pkl", "wb") as pf:
            pickle.dump(paths, pf)
        co_n = cb.opti_cycle_comp_helpers.extract_cb_conf_from_x(opt_res.x, 
                                                                 configs_m, 
                                                                 paths)
        rte, res_ = cb.cb_comp.cb_calc(dir_names_both, POWER_C, config=co_n, plotting=True)
        col_names = cb.opti_cycle_comp_helpers.extract_column_names_from_config(res_["hp"]['output']["config"], paths) + ["rte"]
        for key in res_.keys(): # save plots
            file_fig = dir_res / (current_date + key+"_cb_opti_res.png")
            res_[key]["figure"].savefig(file_fig)
        if HOW_OPT== 'dif_evol':
            res_combi = np.column_stack([opt_res.population, opt_res.population_energies])
            df = pd.DataFrame(res_combi, columns=col_names)
            df.to_csv(file_res, sep=",", index=False)
        else:
            res_dict ={'x': (opt_res.x).tolist(),
                       'fun':float(opt_res.fun),
                       'success':opt_res.success,
                       'message':opt_res.message,
                       'results':dict(zip(col_names, (opt_res.x).tolist()))}
            with res_dat.open('w') as f:
                yaml.dump(res_dict, f, default_flow_style=False,
                           sort_keys=False)
            
        
        


# -*- coding: utf-8 -*-
"""
Created on Thu Aug  8 13:00:55 2024
use Saltelli-sequence to create parameter sample set within defined bounds,
then calls modell including multiprocessing
@author: welp
"""

import carbatpy as cb
from SALib.sample import saltelli
import time
import multiprocessing
import concurrent.futures
import numpy as np
from datetime import datetime
import os
import json

def create_saltelli_sample(dic, n):
    '''
    uses SALib to create saltelli sample

    Parameters
    ----------
    dic : TYPE
        DESCRIPTION. example: 
                    {'num_vars': 5, 
                      'names': ['dT', 'p_ve', 'p_e', 'xa', 'xb'],
                      'bounds': [[2, 25], [2, 8], [200, 600], [0.5, 0.7], [0.05, 0.29]]
                      }   
    n : TYPE
        DESCRIPTION. 2 ** n in Saltelli sequence

    Returns
    -------
    param_values : TYPE
        DESCRIPTION. parameter set

    '''
    N_sample = 2 ** n
    param_values = saltelli.sample(problem=dic, N=N_sample, calc_second_order=False)
    return param_values


def model(args):
    '''
    mock model, use wished model here
    '''
    return np.concatenate((args, [np.sum(args)]))


def process_task(modl, param_values_chunk, path):
    '''
    internal function called within multiprocessing
    '''
    results = []
    for i in range(len(param_values_chunk)):
        param_values = param_values_chunk[i, :]
        res_temp = modl(param_values)
        results.append(res_temp) 
    
    with open(path + '\\Results.txt', 'a') as file:
        for row in range(len(results)):
            file.write(','.join(map(str, results[row])) + '\n')  
    
    
def call_multiprocessing(function, param_set, path='default'):
    '''
    uses multiprocessing to call function with param_set

    Parameters
    ----------
    function : TYPE
        DESCRIPTION. function to be called in multiprocessing
    param_set : TYPE
        DESCRIPTION. parameter set, each row is one parameter set
    path : TYPE, optional
        DESCRIPTION. The default is RES_DIR

    Returns
    -------
    None.

    '''
    s = time.time()
    if path == 'default':
        current_time = datetime.now()
        timestamp = current_time.strftime("%Y-%m-%d_%H-%M-%S")
        path = cb.CB_DEFAULTS["General"]["RES_DIR"] + '\\parameter_sets\\' + timestamp
    os.makedirs(path)
    num_processes = int(multiprocessing.cpu_count() / 2) 
    chunk_size = int(len(param_set) // num_processes)
    
    param_chunks = [param_set[i:i + chunk_size] for i in range(0, len(param_set), chunk_size)]
    with concurrent.futures.ProcessPoolExecutor(max_workers=num_processes) as executor:
        futures = [executor.submit(process_task, function, chunk, path) for chunk in param_chunks]

    e = time.time()
    with open(path + '\\Time.txt', 'a') as f:
        f.write("\nRuntime = {} s ({} h)".format(np.round(e - s, 1), np.round((e - s) / 3600, 2)))

    with open(path + '\\parameters.txt', 'a') as f:
        f.write('input data:\n')
        f.write(json.dumps(input_data, indent=4))  # Use json.dumps for pretty printing
        f.write('\n')
        
        # Write called function
        f.write('called function:\n')
        f.write(str(function.__name__))  # Get the function name
        f.write('\n')
        
    
    
if __name__ == "__main__":
    input_data = {'num_vars': 5, 
                  'names': ['dT', 'p_ve', 'p_e', 'xa', 'xb'],
                  'bounds': [[2, 25], [2, 8], [200, 600], [0.5, 0.7], [0.05, 0.29]]
                  }   
    n = 7
    param_values = create_saltelli_sample(input_data, n)
    call_multiprocessing(model, param_values)
   


"""Top-level package for carbatpy."""

__author__ = """Burak Atakan"""
__email__ = 'burak.atakan@uni-due.de'


# __all__ =[]
import os
import sys
# read version from installed package
from importlib.metadata import version
from importlib.util import find_spec


import pandas as pd

try:
    __version__ = version('carbatpy')
except:
    __version__ = "unknown"

pd.set_option("mode.copy_on_write", True)

try:
    from cb_config import _T_SURROUNDING, _RESULTS_DIR, \
                            _P_SURROUNDING,_CARBATPY_BASE_DIR, TREND, CB_DEFAULTS
   
except:
    from .cb_config import _T_SURROUNDING, _RESULTS_DIR, \
                            _P_SURROUNDING,_CARBATPY_BASE_DIR, TREND, CB_DEFAULTS
 
import carbatpy.models as models
import carbatpy.utils as utils
import carbatpy.helpers as helpers



from carbatpy.utils import exergy_loss as exlo
from carbatpy.utils import cost_calculation as cost
from carbatpy.helpers import file_copy, ser_dict, opti_cycle_comp_helpers

import carbatpy.models.cb_fluids.fluid_props as fprop
from carbatpy.models.cb_fluids.fluid_props import Fluid, FluidModel, FluidState, init_fluid
#from carbatpy.models.components import compressor_simple, throttle_simple
from carbatpy.models.components import heat_exchanger_thermo_v2 as hex_th
from carbatpy.models.components.single_flow import FlowDeviceOld
from carbatpy.models.components import comp
from carbatpy.models.components.class_he import  heat_exchanger
from carbatpy.models.components.surrogates import  Surrogate
import carbatpy.models.components.heat_transfer 


from carbatpy.models.coupled import heat_pump_comp as hp_comp
from carbatpy.models.coupled import orc_comp, cb_comp
from carbatpy.models.coupled import heat_pump_simple as hp_simple
from carbatpy.models.coupled import orc_simple_v2 as orc_simple
from carbatpy.models.coupled import read_cycle_structure
from carbatpy.utils import curve_min_distance_finder, property_eval_mixture, optimize

HAS_SPP_MACHINES = find_spec("spp_machines") is not None
#cost_inst = cost.CAP_methods()  # Instance of cost class, shall be instantiated only once

_cost_inst = None

def get_cost_inst():
    """Return global CAP_methods instance (lazy initialization)."""
    global _cost_inst
    if _cost_inst is None:
        from .utils import cost_calculation as cost
        _cost_inst = cost.CAP_methods()
    return _cost_inst



# -*- coding: utf-8 -*-
"""
Script to read the content of a Carnot Battery Optimization directory

and do the CB calcultions with the optimal results and plot them.

Created on Fri Aug  8 16:07:33 2025

@author: atakan
Universität Duisburg-Essen, Germany

In the framework of the Priority Programme: "Carnot Batteries: Inverse Design from
Markets to Molecules" (SPP 2403)
https://www.uni-due.de/spp2403/
https://git.uni-due.de/spp-2403/residuals_weather_storage

"""

import numpy as np
import pandas as pd
import carbatpy as cb
from pathlib import Path
import yaml
import pickle

# === EINSTELLUNGEN ===
# Replace the results directory with the one wanted:
RES_DIR = Path(r"C:\Users\atakan\sciebo\results\2025-08-08-17-07-cb_opt_result")  # Ordner vom ersten Skript
CSV_FILE = list(RES_DIR.glob("*cb_opti_res.csv"))[0]  # nimmt automatisch die einzige CSV
YAML_FILE = RES_DIR / "config_bound.yaml"

PLOT = True        # Plots speichern?
ALL_POP = False    # True = alle Zeilen aus CSV berechnen, False = nur beste Lösung

# === DATEIEN LADEN ===
with open(YAML_FILE, "r") as f:
    yaml_data = yaml.safe_load(f)

configs_m = yaml_data["configs"]
bounds_m = yaml_data["bounds"]
POWER_C = yaml_data["power_compressor"]


# io-Dateien
dir_names_both = {
    "hp": str(RES_DIR / "io-cycle-data.yaml"),
    "orc": str(RES_DIR / "io-orc-data.yaml"),
}


# === Zuerst versuchen, saved paths zu laden ===
paths_file = RES_DIR / "paths.pkl"
if paths_file.exists():
    with open(paths_file, "rb") as pf:
        paths = pickle.load(pf)
    print("Loaded 'paths' from paths.pkl")
else:
    # Fallback: selbst erzeugen (verwende Deine Funktion)
    print("No 'paths.pkl' found — reconstructing paths from config/bounds (best-effort).")
    paths = {}
    for k in ["hp", "orc"]:
        _, _, p_list = cb.opti_cycle_comp_helpers.extract_optim_data_with_paths(configs_m[k], bounds_m[k])
        paths[k] = p_list

# === CSV lesen (Index-Spalte ausschließen) ===
df = pd.read_csv(CSV_FILE, index_col=0)  # falls Optimierer index=False nimmt, ist das auch ok
energy_col = df.columns[-1]
x_cols = df.columns[:-1].tolist()

# === Sanity-check: Anzahl x in paths vs CSV ===
expected_len = sum(len(paths[k]) for k in ("hp", "orc"))
if expected_len != len(x_cols):
    raise RuntimeError(
        f"Mismatch between expected number of x-variables ({expected_len}) from 'paths' and CSV columns ({len(x_cols)}). "
        f"CSV columns (first 12): {x_cols[:12]}. Save 'paths.pkl' during optimization or check CSV writing."
    )
print(f"OK: expected_len={expected_len}, csv_x_cols={len(x_cols)}")

# === calc-Funktion ===
def calc_from_row(row, idx):
    x_vals = row[x_cols].values.astype(float)
    co_n = cb.opti_cycle_comp_helpers.extract_cb_conf_from_x(x_vals, configs_m, paths)
    rte, res_ = cb.cb_comp.cb_calc(dir_names_both, POWER_C, config=co_n, plotting=PLOT)
    if PLOT:
        for key in res_:
            file_fig = RES_DIR / f"recalc_{idx}_{key}.png"
            res_[key]["figure"].savefig(file_fig)
    return rte

# === Auswahl ===
if ALL_POP:
    results = []
    for i, row in df.iterrows():
        rte_val = calc_from_row(row, i)
        results.append((i, rte_val))
    print("Alle Ergebnisse neu berechnet:", results)
else:
    best_idx = df[energy_col].idxmin()
    rte_val = calc_from_row(df.loc[best_idx], best_idx)
    print(f"Beste Lösung neu berechnet (Index {best_idx}): RTE = {rte_val}")



# -*- coding: utf-8 -*-
"""
Created on Sun Jan  4 11:28:38 2026

@author: atakan
Universität Duisburg-Essen, Germany

In the framework of the Priority Programme: "Carnot Batteries: Inverse Design from
Markets to Molecules" (SPP 2403)
https://www.uni-due.de/spp2403/
https://git.uni-due.de/spp-2403/residuals_weather_storage

"""

import copy
from types import SimpleNamespace
import numpy as np
import matplotlib.pyplot as plt
import carbatpy as cb


def orc(dir_name, cop, q_dot_high, **kwargs):
    """
    Compute an Organic Rankine Cycle (ORC) for a Carnot battery.

    The ORC uses an isolated configuration (from file or dict). The evaporator heat
    input q_dot_high and the COP of the preceding charging heat pump are prescribed;
    the working fluid mass flow is determined accordingly. Components: pump, evaporator,
    expander, condenser to ambient (condenser_sur), and main condenser.

    Key properties
    - No in-place modifications to CB_DEFAULTS or externally provided configurations.
    - Configurations are handled in isolation (deepcopy/recursive merges).
    - Matplotlib Figure/Axes are consistently passed to all plot calls so that the
      entire cycle is drawn on the same axes.

    Parameters
    ----------
    dir_name : str or dict
        Path to a YAML configuration file or a configuration dict.
    cop : float
        COP of the preceding charging heat pump.
    q_dot_high : float
        Prescribed heat flow rate to the evaporator [W].
    config : dict, optional (via kwargs)
        Overrides to the base configuration (e.g., for optimization).
    verbose : bool, optional (via kwargs)
        If True, print additional information.
    plotting : bool, optional (via kwargs)
        If True, create a combined cycle plot.

    Returns
    -------
    dict
        Result dictionary with keys:
        - "eta_th" : float
            Thermal efficiency (net electric power / evaporator heat input).
        - "output" : dict
            Component outputs and the configuration used ("config").
        - "warnings" : dict
            Warnings per component (and additional items, e.g., pressure_ratio).
        - "figure" : matplotlib.figure.Figure or None
            Figure object, if plotting=True.
        - "axes" : matplotlib.axes.Axes or None
            Axes object, if plotting=True.
        - "costs" : any
            Aggregate cost metric from components (from cb.orc_comp.all_costs).

    Raises
    ------
    KeyError
        If required configuration entries are missing (e.g., working_fluid.p_high).
    """
    new_config = kwargs.get("config", None)
    verbose = kwargs.get("verbose", False)
    plotting = kwargs.get("plotting", False)

    fig = None
    ax = None
    warnings = {}
    outputs = {}

    def add_w_o(comp):
        # collect warnings/outputs per component
        warnings[comp.name] = copy.deepcopy(comp.warning)
        outputs[comp.name] = copy.deepcopy(comp.output)

    # Isolated, merged configuration
    config = cb.hp_comp.resolve_config(dir_name, new_config)

    # Initial start (mass flow will be set later)
    start0 = cb.comp.Start("start", config, m_dot=10e-3, verbose=verbose)
    p_low = config["working_fluid"]["p_low"]
    p_high = config["working_fluid"]["p_high"]

    # Prescribed heat flows (sign convention: negative for heat rejected)
    run_p_evap = {"q_dot": q_dot_high}
    q_dot_loss = -q_dot_high * (1.0 - 1.0 / cop)  # heat to cold storage (negative)
    run_p_cond = {"q_dot": q_dot_loss}

    # Pump (first pass with m_dot=1 to set states)
    pump = cb.comp.FlowMachine("pump", config, verbose=verbose)
    p_ratio = p_high / start0.output["state_out"]["working_fluid"][1]
    if p_ratio < 1.0:
        warnings["pressure_ratio"] = SimpleNamespace(
            value=10.0 / p_ratio,
            message="Pressure ratio is wrong"
        )

    pump.calculate(
        start0.output["state_out"],
        {"working_fluid": [600, p_high, 5e5]},
        run_param={"m_dot": {"working_fluid": 1.0}}
    )

    # Evaporator: determines actual m_dot via q_dot_high
    evaporator = cb.comp.StaticHeatExchanger("evaporator", config, verbose=verbose)
    inp_evap, _ = evaporator.set_in_out({"working_fluid": pump.output["state_out"]["working_fluid"]})
    evaporator.calculate(inp_evap, run_param=run_p_evap, verbose=verbose)
    volumes_e = evaporator.calculate_volume()
    add_w_o(evaporator)
    if verbose:
        print(f"Storage Volumes evaporator: {volumes_e}")

    # Correct start with determined m_dot
    m_dot_w = evaporator.output["m_dot"]["working_fluid"]
    start = cb.comp.Start("start", config, m_dot=m_dot_w, verbose=verbose)
    run_p_machine = {"m_dot": {"working_fluid": m_dot_w}}

    # Recompute pump with correct m_dot (if needed by your component implementation)
    pump.calculate(
        start.output["state_out"],
        {"working_fluid": [600, p_high, 5e5]},
        run_param=run_p_machine,
        m_dot=m_dot_w
    )
    add_w_o(start)
    add_w_o(pump)

    # Expander
    expander = cb.comp.FlowMachine("expander", config, verbose=verbose)
    expander.calculate(
        evaporator.output["state_out"],
        {"working_fluid": [600, p_low, 5e5]},
        m_dot=m_dot_w,
        run_param=run_p_machine,
        verbose=verbose
    )
    add_w_o(expander)

    # Net power
    power_tot = expander.output["power"] + pump.output["power"]

    # Condenser to surroundings (heat not going into storage)
    q_dot_sur = -(q_dot_high + q_dot_loss + power_tot)  # choose sign so run_param is positive
    run_p_cond_sur = {"q_dot": -q_dot_sur, "m_dot": {"working_fluid": m_dot_w}}

    condenser_sur = cb.comp.StaticHeatExchanger("condenser_sur", config, verbose=verbose)
    inp_cs, _ = condenser_sur.set_in_out({"working_fluid": expander.output["state_out"]["working_fluid"]})
    condenser_sur.calculate(inp_cs, run_param=run_p_cond_sur, verbose=verbose)
    add_w_o(condenser_sur)

    # Main condenser (to cold storage)
    run_p_cond.update({"q_dot": -q_dot_loss, "m_dot": {"working_fluid": m_dot_w}})
    condenser = cb.comp.StaticHeatExchanger("condenser", config, verbose=verbose)
    inp_c, _ = condenser.set_in_out({"working_fluid": condenser_sur.output["state_out"]["working_fluid"]})
    sec_in, sec_out = condenser.set_in_out({"working_fluid": start.output["state_in"]["working_fluid"]}, False)
    condenser.calculate(inp_c, sec_out, run_param=run_p_cond, verbose=verbose)
    volumes_c = condenser.calculate_volume()
    add_w_o(condenser)
    if verbose:
        print(f"Storage Volumes condenser: {volumes_c}")

    # Efficiency and costs
    eta_th = np.abs(power_tot) / np.abs(evaporator.output["q_dot"])
    costs = cb.orc_comp.all_costs([pump, evaporator, expander, condenser_sur, condenser],
                                  verbose)

    if verbose:
        print(f"eta_th: {eta_th:.4f}")

    # Plot: shared Figure/Axes, do not deepcopy fig/ax
    if plotting:
        fig, ax = plt.subplots(1)
        # Copy plot defaults (pure values), then attach fig/ax
        plot_info = copy.deepcopy(cb.CB_DEFAULTS["Components"]["Plot"])
        plot_info["ax"] = ax
        plot_info["fig"] = fig
        plot_info["x-shift"] = [0, 0]

        # Start point on same axes: use shallow copy so fig/ax references are preserved
        start_plot_info = plot_info.copy()
        start_plot_info.update({
            "label": ["start", ""],
            "col": ["ok", "bv"],
            "direction": 1
        })
        shift, direct = start.plot(start_plot_info)

        # Subsequent components on the same axes
        plot_info["x-shift"] = shift
        plot_info["direction"] = direct

        plot_info.update({"label": [pump.name, ""], "col": ["-r", "bv-"]})
        shift, direct = pump.plot(plot_info)

        plot_info["x-shift"] = shift
        plot_info["direction"] = direct
        plot_info.update({"label": [evaporator.name, ""], "col": [":r", "rv-"]})
        shift, direct = evaporator.plot(plot_info)

        plot_info["x-shift"] = shift
        plot_info["direction"] = direct
        plot_info.update({"label": [expander.name, ""], "col": ["-b", "bv-"]})
        shift, direct = expander.plot(plot_info)

        plot_info["x-shift"] = shift
        plot_info["direction"] = direct
        plot_info.update({"label": [condenser_sur.name, ""], "col": [":b", "bv-"]})
        shift, direct = condenser_sur.plot(plot_info)

        plot_info["x-shift"] = shift
        plot_info["direction"] = direct
        plot_info.update({"label": [condenser.name, ""], "col": [":r", "rv-"]})
        condenser.plot(plot_info)

    # Collect results
    outputs["config"] = copy.deepcopy(config)
    results = {
        "eta_th": eta_th,
        "output": outputs,
        "warnings": warnings,
        "figure": fig,
        "axes": ax,
        "costs": costs
    }
    return results


def all_costs(components_all, verbose):
    cost_total = 0
    cost_dict ={}

    for proc in components_all:
        cost = proc.estimate_costs()
        cost_total += cost
        cost_dict[proc.name] = cost
        if 'Storage_costs' in proc.output.keys():
            for key, val in proc.output['Storage_costs'].items():
                cost_dict[proc.name+"-storage-" + key] = val
                cost_total +=val
                
        if verbose:
            print(f'{proc.name} --Exergy destr.rate: {proc.output["exergy_destruction_rate"]:.2e} W, costs: {proc.cost:.2e}')
            if 'Storage_costs' in proc.output.keys():
                print(f"\tStorage costs: {proc.output['Storage_costs']}")
    return {"total_costs":cost_total, "cost_distribution":cost_dict}


if __name__ == "__main__":
    dir_name_m = cb.CB_DEFAULTS["General"]["CB_DATA"]+"\\io-orc-data.yaml"
    conf_m = {"cold_storage": {"temp_low": 250.},
              "working_fluid": {"p_low": 7e5,
                                "p_high": 11e5,
                                'species': 'ethane * propane * pentane * hexane',
                                'fractions': [0.25,
                                              0.55,
                                              0.15,
                                              0.05]}}
    res_act = orc(dir_name_m, 3, 1000, config=conf_m,
                                         plotting =True,
                                         verbose=True)
    if any(ns.value != 0 for ns in res_act["warnings"].values()):
        print(f"Check warnings, at least one deviates from 0!\n{res_act['warnings']}")
    


# -*- coding: utf-8 -*-
"""

This script optimizes the pressure of the working fluid in a heat exchanger

the mean temperature difference is minimized, without having a point below
the minimum approach temperature (pinch point).
it works, but is slow.

Created on Tue Jul 23 11:26:13 2024

@author: atakan
"""


import copy
from time import time
# import src.models.fluids.fluid_props as fprop
import numpy as np
from scipy.optimize import minimize, minimize_scalar, Bounds  # root, root_scalar
import matplotlib.pyplot as plt
import carbatpy as cb


FLUID = "Propane * Pentane"  # working fluid
FLS = "Methanol"  # "Water"  # secondary fluid
comp = [.50, 0.5]
flm = cb.fprop.FluidModel(FLUID)
myFluid = cb.fprop.Fluid(flm, comp)

secFlm = cb.fprop.FluidModel(FLS)
secFluid = cb.fprop.Fluid(secFlm, [1.])
D_TEMP_MIN = 5.0

# Condenser, working fluid fixes all, secondary output enthalpy can be varied:
SEC_TEMP_IN = 300.0
SEC_TEMP_OUT_MAX = 370.0
SEC_PRES_IN = 5e5
H_DOT = 1e3
state_sec_out = secFluid.set_state([SEC_TEMP_OUT_MAX, SEC_PRES_IN], "TP")

state_sec_in = secFluid.set_state(
    [SEC_TEMP_IN, SEC_PRES_IN], "TP")  # this is the entering state

# working fluid

TEMP_SAT_VAP = SEC_TEMP_OUT_MAX + D_TEMP_MIN
state_in = myFluid.set_state(
    [TEMP_SAT_VAP, 1.], "TQ")  # find minimum pressure

WF_TEMP_IN = TEMP_SAT_VAP + D_TEMP_MIN
WF_TEMP_OUT = SEC_TEMP_IN + D_TEMP_MIN
state_out = myFluid.set_state([WF_TEMP_OUT, state_in[1]], "TP")

# now plotting can directly be done in pinch_calc 2024-05-24
fig_act, ax_act = plt.subplots(1)
PLOT_INFO = {"fig": fig_act, "ax": ax_act, "what": [2, 0], "col": ["r:", "ko"],
             "label": ["work,c", "sec,c"], "x-shift": [0, 0]}

# a simple way to find an optimal/better pressure level
pressures_good =[]
p_start = state_out[1]

# p_range = np.linspace(p_start*.87, p_start*1.015,11)
# for p_act in p_range:

#     state_in = myFluid.set_state([p_act,
#                                   WF_TEMP_IN],
#                                  "PT")
#     # myFluid.print_state()

#     hex0 = cb.models.components.heat_exchanger_thermo_v2.StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
#                                state_sec_out[2],
#                                d_temp_separation_min=D_TEMP_MIN)
    

#     factor0 = hex0.find_pinch()
#     if hex0.warning ==0:
#         pressures_good.append([p_act, hex0.dt_mean, hex0.warning, hex0.warning_message])
#     print(p_act, hex0.dt_mean, hex0.warning, hex0.warning_message)
# if hex0.warning > 0:
#     print(hex0.warning_message)
# print("useful pressures:\n",pressures_good,"\n\n")    
# hex0 = cb.models.components.heat_exchanger_thermo_v2.StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
#                             state_sec_out[2],
#                             d_temp_separation_min=D_TEMP_MIN)

def to_opt(p_act):
    state_in = myFluid.set_state([p_act,
                                  WF_TEMP_IN],
                                 "PT")
    # myFluid.print_state()

    hex_act = cb.models.components.heat_exchanger_thermo_v2.StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
                                state_sec_out[2],
                                d_temp_separation_min=D_TEMP_MIN)
    

    factor0 = hex_act.find_pinch()
    if hex_act.warning == 0:
        return hex_act.dt_mean
    return 80*hex_act.dt_mean

bound_act =Bounds(lb =[p_start*.75], ub =[p_start *1.1])
tolerance = 5e-1
t0 = time()
res_scalar = minimize_scalar(to_opt,
                             bounds =[.75*p_start, 1.1*p_start],
                             #tol=tolerance,
                             options ={"maxiter": 14,
                                       "disp": True,
                                       "xatol":tolerance}
                             )
print(res_scalar)
t1 =time()
result = minimize(to_opt,p_start,
                  method='Nelder-Mead',
                  tol=tolerance,
                  bounds=bound_act,
                  options ={"maxiter": 14,
                            "disp": True})
print(result)
t2 =time()
result = minimize(to_opt,p_start,
                  method='L-BFGS-B',
                  tol=tolerance,
                  bounds=bound_act,
                  options ={"maxiter": 14,
                            "disp": True})
t3 =time()
print(result)
print(t1-t0, t2-t1, t3-t2)



# -*- coding: utf-8 -*-
"""

Created on Sun Aug  4 09:44:00 2024

@author: atakan
"""


import carbatpy as cb
import json
import yaml


def read_config(file_path):
    if isinstance(file_path,dict):
        return file_path

    if file_path.endswith('.yaml') or file_path.endswith('.yml'):
        return read_yaml(file_path)
    elif file_path.endswith('.json'):
        return read_json(file_path)
    else:
        raise ValueError("Unsupported file format. Use .yaml, .yml, or .json")


def read_yaml(file_path):
    import yaml
    with open(file_path, 'r') as file:
        return yaml.safe_load(file)


def read_json(file_path):
    with open(file_path, 'r') as file:
        return json.load(file)


def fluid_from_dict_file(fluid_name, path_dict, **kwargs):
    info = path_dict
    if isinstance(path_dict, str):
        info = read_config(path_dict)
    fluid_act = info.get(fluid_name)

    fl_instance = cb.init_fluid(fluid_act["species"],
                                fluid_act["fractions"],
                                props=fluid_act["props"])
    return fl_instance


def read_component(config_dict, name=None, temp_ambient=None):
    parsed_components = {}
    if isinstance(config_dict, str):
        config_dict = read_config(config_dict)
    if name:
        parsed_component = parse_equipment(
            name, config_dict[name], config_dict, temp_ambient)
        return parsed_component
    else:

        for name, config in config_dict.items():
            if "model" in config:
                parsed_components[name] = parse_equipment(
                    name, config, config_dict, temp_ambient)
        return parsed_components


def parse_equipment(name, config, config_all, temp_ambient):
    model = config['model']
    parameters = config.get('parameters')
    calc_type = config.get('calc_type')
    species = config.get('species', {})
    dtemp_min = config.get("dt_min")
    if "low" in species['working_fluid']['in']:
        dtemp_min = -parameters["dt_min"]
    states = {}

    act_fluids = {}

    for fluid, flow in species.items():
        act_fluids[fluid] = fluid_from_dict_file(fluid, config_all)
        fl = config_all[fluid]
        states[fluid] = {}
        for where in ["out", "in"]:
            # if not temperatures are prescribed, but other components, no states are calculated
            if "temp" in flow[where]:
                if flow[where] =="temp_start": # determine an initial state, if not given. Changes/sets pressure!
                    states[fluid][where]  = _get_start(config_all, 
                                                       act_fluids[fluid],
                                                       species,
                                                       temp_ambient)
                    
                # BA 2025-07-28: second posibility "temp_start_tp" for fixing temp and p? Superheating must then be checked independently. Could be better for optimizations.    
                else:
                    temp = fl[flow[where]] if where in flow else None
                    if temp == 'ambient':
                        temp = cb.CB_DEFAULTS["General"]["T_SUR"]
                        if temp_ambient:
                            temp = temp_ambient
                    if where == "out" and "working" in fluid:
                        temp += dtemp_min  # BA looks dangerous 2024-08-15
    
                    states[fluid][where] = act_fluids[fluid].set_state(
                            [temp, fl[flow["p_" + where]]], "TP")
            else:
                states[fluid][where] = flow[where]
                # will be calculated along cycle

        
        

    return {
        'name': name,
        'model': model,
        'calc_type': calc_type,
        'states': states,
        'act_fluids': act_fluids,
        'parameters': {k: v for k, v in config.items() if k not in ['model', 'calc_type', 'species']}
    }


def _get_start(config, fluid_act, which, temp_ambient):
    """
    Calculate starting points for cycles, when not given otherwise.
    
    Only for heat pumps so far. Evaporator outlet should be superheated by a given
    value and have a minimum distance to the entering storage fluid.
    
    

    Parameters
    ----------
    config : dictionary
        the whole cycle configuration.
    fluid_act : Fluid
        the working fluid (model) to set its state. This input state is changed!
    which : dictionary
        with both fluid-species names, to find the non-working-fluid.
    temp_ambient : TYPE
        if this is the value of the storage.

    Raises
    ------
    NotImplementedError
        DESCRIPTION.

    Returns
    -------
    FluidState
        the state of the starting point.
    state_sh : TYPE
        DESCRIPTION.

    """
    if config["process"]["name"]=="heat_pump":
        keys = list(which.keys()  ) 
        for key in keys:
            if key != 'working_fluid':
                other_key = key
                break
        temp_key = which[other_key]["in"]
        temp =  _set_temp(config[other_key][temp_key], temp_ambient)
        temp_sat =temp - config["start"]["dt_min"]- config["start"]["dt_superheat"]
        
        sat = fluid_act.set_state([temp_sat,1],"TQ")
        temp_sh = temp - config["start"]["dt_min"]
        n_pres =cb.CB_DEFAULTS["Fluid_Defaults"]['Property_Names'].index("Pressure")
        
        state_sh = fluid_act.set_state([temp_sh, sat[n_pres]],"TP")
        return state_sh 
    
    elif config["process"]["name"].lower() in ("organic_rankine_cycle", "orc"):
        keys = list(which.keys()  ) 
        for key in keys:
            if key != 'working_fluid':
                other_key = key
                break
        temp_key = which[other_key]["in"]
        temp =  _set_temp(config[other_key][temp_key], temp_ambient)
        temp_sat = temp + config["start"]["dt_min"] + config["start"]["dt_subcool"]
        p_low =config["working_fluid"]["p_low"]
        
        sat = fluid_act.set_state([p_low,0],"PQ")
        temp_sh =sat[0] -  config["start"]["dt_subcool"]
        if temp_sh < temp_sat:
            temp_sh = temp + config["start"]["dt_min"]
        n_pres = cb.CB_DEFAULTS["Fluid_Defaults"]['Property_Names'].index("Pressure")
        
        state_sh = fluid_act.set_state([temp_sh, sat[n_pres]],"TP")
        return state_sh 
        
    else:
        raise NotImplementedError("other cycles are not implemented yet (starting point)")
        
        
def _set_temp(temp, temp_ambient =None):
    """
    Set the ambient temperature when temp ='ambient'.
    

    Parameters
    ----------
    temp : TYPE
        DESCRIPTION.
    temp_ambient : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    temp : TYPE
        DESCRIPTION.

    """
    if temp == 'ambient':
        temp = cb.CB_DEFAULTS["General"]["T_SUR"]
        if temp_ambient:
            temp = temp_ambient
    return temp


def setup_system(config_path):
    config = config_path
    if isinstance(config_path, str):

        config = read_config(config_path)

    # Komponenten einlesen
    components = read_component(config)

    # Fluide einlesen
    fluids = {}
    for name, fluid_config in config.items():
        if "props" in fluid_config:
            fluids[name] = fluid_from_dict_file(name, config)

    return components, fluids


if __name__ == "__main__":
    input_hex = {
        'process': {"cycle": ["start",
                              "compressor",
                              "condenser",
                              "throttle",
                              "evaporator",
                              "start"],
                    "name": "heat_pump",
                    },

        'fluids_all': {'working_fluid': 'cycle',
                       'cold_storage': 'source-sink',
                       'hot_storage': 'source-sink'
                       },
        'start': {
            'model': "Start",
            "calc_type": None,
            "species": {"working_fluid": {"in": "temp_start",
                                          "out": "temp_start",
                                          "p_in": "p_low",
                                          "p_out": "p_low"},
                        "cold_storage": {"in": "temp_high",
                                        "out": "temp_high",
                                        "p_in": "p_low",
                                        "p_out":"p_low"}},
            "dt_min": 3.,
            "dt_superheat": 5.,
            
        },

        'compressor': {
            'model': "FlowDevice",
            'calc_type': 'const_eta',
            'eta_s': 0.7,
            'fixed': 'power',
            'power': 1000.,
            'species': {'working_fluid': {'in': 'evaporator',
                                          'out': 'condenser',
                                          "p_in": "p_low",
                                          "p_out": "p_high"}},
            'dt_superheat': 5
        },

        'throttle': {
            'model': "FlowDevice",
            'calc_type': 'const_h',
            'fixed': "m_dot",
            'species': {'working_fluid': {'in': 'condenser',
                                          'out': 'evaporator',
                                          "p_in": "p_high",
                                          "p_out": "p_low"}}
        },



        'condenser': {
            'model': "StaticHeatExchanger",
            "calc_type": "const",
            'fixed': "m_dot_w",
            "species": {"working_fluid": {"in": "temp_high",
                                          "out": "temp_low",
                                          "p_in": "p_high",
                                          "p_out": "p_high"},
                        "hot_storage": {"in": "temp_low",
                                        "out": "temp_high",
                                        "p_in": "p_low",
                                        "p_out":"p_low"}},
            "dt_min": 3.,
            "q_dot": 1000.,
            "overall_u": 100,
            'area': 1,
        },

        'evaporator': {
            'model': "StaticHeatExchanger",
            'calc_type': 'const',
            'fixed': "m_dot_w",
            'species': {'working_fluid': {'in': 'throttle',
                                          'out': 'compressor',
                                          "p_in": "p_low",
                                          "p_out":"p_low"},
                        'cold_storage': {'in': 'temp_high',
                                         'out': 'temp_low',
                                         "p_in": "p_low",
                                         "p_out":"p_low"}},
            'dt_min': 3.0,
            
            "overall_u": 100,
            'area': 1,
        },

        'hot_storage': {
            "species": "Water",
            "fractions": [1.0],
            "p_low": 2e5,
            "temp_low": "ambient",
            "temp_high": 370.0,
            'props': "REFPROP",
            
            'fixed': None,
        },


        'cold_storage': {
            "species": "Methanol",
            "fractions": [1.0],
            "p_low": 2e5,
            "temp_low": 255.0,
            "temp_high": "ambient",
            'props': "REFPROP",
            'fixed': None,
        },
        'working_fluid': {
            "species": "Propane * Butane * Pentane * Hexane",
            "fractions": [
                7.71733787e-01,
                2.22759488e-02,
                1.78685867e-01,
                0.027304397199999997],
            "temp_low": "ambient",
            "temp_high": 420.0,
            "p_low": 1.128250708e+05,
            "p_high": 1.37548728e+06,
            "optimize": "None",
            "setting": "initial",
            'props': "REFPROP",
        },
    }

    act_fl = fluid_from_dict_file("working_fluid", input_hex)
    act_fl.set_state([300, 1e5], "TP")
    act_fl.print_state()
    # aa = setup_system(input_hex)
    # print("\n", aa)
    start = read_component(input_hex, "start")
    compressor = read_component(input_hex, "compressor")
    condenser = read_component(input_hex, "condenser")
    p_names =cb.CB_DEFAULTS["Fluid_Defaults"]['Property_Names']
    n_temp= p_names.index("Temperature")
    n_pres =p_names.index("Pressure")
    print( condenser["states"]["working_fluid"]["in"][n_temp],
          condenser["states"]["working_fluid"]["out"][n_temp],
          condenser["states"]["working_fluid"]["in"][n_pres])
    fn = cb.CB_DEFAULTS["General"]["RES_DIR"]+"\\io-cycle-data2.yaml"
    
    with open(fn, "w") as file:
        yaml.safe_dump(input_hex, file)



# -*- coding: utf-8 -*-
"""
Calculating costs for equipment

SI units input is only implemented for the Towler method and the limits are
checked now for Towler! (BA 2025-08-14/15)

Created on Wed Jul 31 13:59:25 2024

@author: Mina Anton
"""

import numpy as np
import pandas as pd
import datetime
import sys
import warnings
import carbatpy as cb


# Pfad zur Excel-Datei
DIR = cb.CB_DEFAULTS["General"]['CB_DATA']
_EXCEL_FILE = DIR + "\\" + "Capital_Investment_Methods_ba.xlsx"

# Speicherstrukturen: Dict mit {sheetname: {"columns": [...], "values": np.array([...])}}
_CAP_DATA = {}

try:
    xl_data = pd.ExcelFile(_EXCEL_FILE, engine="openpyxl")

    for sheet in ["Couper Method", "Turton Method", "Towler Method", "CEPCI data"]:
        df = xl_data.parse(sheet)
        _CAP_DATA[sheet] = {
            "columns": list(df.columns),
            "values": df.values,
            "df": df,
        }
    cols = _CAP_DATA["Towler Method"]["columns"]
    _Towler_unit_col_indices = [
        i for i, col in enumerate(cols)
        if str(col).lower().startswith("unit")]

except Exception as e:
    print(f"Error loading Excel file: {e}")


class CAP_methods:
    def __init__(self):

        # Direktes Zuweisen der vorbereiteten Daten
        self.Couper_columns = _CAP_DATA["Couper Method"]["columns"]
        self.Couper_xl_data = _CAP_DATA["Couper Method"]["values"]

        self.Turton_columns = _CAP_DATA["Turton Method"]["columns"]
        self.Turton_xl_data = _CAP_DATA["Turton Method"]["values"]

        self.Towler_columns = _CAP_DATA["Towler Method"]["columns"]
        self.Towler_xl_data = _CAP_DATA["Towler Method"]["values"]
        self.Towler_df = _CAP_DATA["Towler Method"]["df"]
        self.Towler_unit_col_index = _Towler_unit_col_indices[0]

        self.CEPCI_columns = _CAP_DATA["CEPCI data"]["columns"]
        self.CEPCI_xl_data = _CAP_DATA["CEPCI data"]["values"]
        self.warn = {'value': 0,
                     'message': 'All o.k.'}

    def _Find_Present_Cost(self, BaseCost, BaseYear, plot=False, Current_year=True, Desired_year=None):
        """
        Function used to adjust the cost of a component to the present day, accounting for inflation. It uses the
        CEPCI (Chemical Engineering cost index data).

        Parameters
        ----------
        BaseCost : TYPE, Float.
            DESCRIPTION: Cost of the compoenent based on the year the correlation is published.
        BaseYear : TYPE, Float.
            DESCRIPTION: Year the correlation for the cost estimate is published.
        plot : TYPE, optional.
            DESCRIPTION. The default is True. Plots the CEPCI vs year.
        Current_year : TYPE, Boolean, optional.
            DESCRIPTION : The default is True. Should the CEPCI cost index be taken relative to the current year.
        Desired_year : TYPE, Float, optional.
            DESCRIPTION : The default is None. If you choose False for Current_year, what year do you want
            the cost index to be relative to.

        Returns
        -------
        costPresent : TYPE, Float.
            DESCRIPTION: Present day cost of the component, accounting for inflation.

        """

        # Finds the year for todays date
        if Current_year == True:
            year = datetime.datetime.now().year
        else:
            if Desired_year is not None:
                year = Desired_year
            else:
                raise ValueError(
                    "Desired year is not provided although Current year was set to False")

        # Find the current and past cost index for the component
        if year in self.CEPCI_xl_data[:, 0]:
            index_present = self.CEPCI_xl_data[np.where(
                self.CEPCI_xl_data[:, 0] == year), 1]
            index_data = self.CEPCI_xl_data[np.where(
                self.CEPCI_xl_data[:, 0] == BaseYear), 1]
        # Check if present date has a cost index in CECPI
        else:
            print("\nerror: no cost index data for ", year)
            sys.exit()

        if plot:
            print("\ncost index ratio: ", index_present[0][0], "/", index_data[0]
                  [0], " = ", np.round((index_present/index_data)[0][0], 2), "\n")

        costPresent = float(BaseCost*(index_present[0][0]/index_data[0][0]))

        return costPresent

    def _Couper_Method(self, Comp_dict, verbose=False, Desired_year=None):
        """
        Method based on cost estimation techniques for indivisual components
        from chapter 21 in Chemical Process
        Equipment, Third Edition (2012).
        
        SI-unit input not implemented yet! Limits are not checked! BA 2025-08-15

        Parameters
        ----------
        Comp_dict : TYPE, Dictionary.
            DESCRIPTION : This dictionary contains all the variables and paramters needed to
            find a cost estimate for a particular component. For the exact values in the dictionary
            check the excel file Capital_Investment_Methods or read chapter 21 in Chemical Process
            Equipment, Third Edition (2012).
            Comp_dict_example = {
            "Category": ,
            "Component Name": ,
            "Power" :,      # For the turton method the pressures are in guage, 100 KPa = 1atm
            "Area": ,
            "Pressure": ,
            "Volume": ,
            "Volume Flow Rate": ,
            "Head":
                }
        verbose : TYPE, Boolean, optional.
            DESCRIPTION : should values be printed along with an explanation?
        Desired_year : TYPE, Float, optional.
            DESCRIPTION : The default is None. Do you want the cost index to be taken relative to the current year or another previous years?
            If you want the cost index to be taken relative to a previuos year, input value to Desired_year.

        Raises
        ------
        ValueError.
            DESCRIPTION : Checks if the component name is available in the Couper method excel sheet.

        Returns
        -------
        TYPE, String.
            DESCRIPTION : The installed cost of the component, adjusted to present day to account
            for inflation. This cost estimate also accounts for operational pressure and material of construction.

        """

        self.Component_category = Comp_dict.get("Category")
        self.Component_name = Comp_dict.get("Component Name")
        self.Power = Comp_dict.get("Power")
        self.Area = Comp_dict.get("Area")
        self.Pressure = Comp_dict.get("Pressure")
        self.Volume = Comp_dict.get("Volume")
        self.Q = Comp_dict.get("Volume Flow Rate")
        self.h = Comp_dict.get("Head")

        # Finds row corresponding to component and checks if component is available in this method
        index = np.where(self.Couper_xl_data[:, 1] == self.Component_name)[0]
        if index.size == 0:
            raise ValueError(
                "Component name is not found in the Couper method")

        # Array index contains the row for the component
        # Finds values needed for the cost estimation for the component
        Component_row = self.Couper_xl_data[index[0]]
        a = float(Component_row[2])
        b1 = float(Component_row[3])
        b2 = float(Component_row[4])
        b3 = float(Component_row[5])
        n = float(Component_row[6])
        iF = float(Component_row[7])
        fm = float(Component_row[8])
        ccF = float(Component_row[9])
        cF = float(Component_row[13])
        cF2 = float(Component_row[17])
        yr = float(Component_row[18])
        fd = None
        fp = None

        # Calculate cost of compressor
        if self.Component_category == "Compressor":

            HP = self.Power * cF
            CI = a * (HP**n) * iF * ccF * 1000
            CI = self._Find_Present_Cost(CI, yr, plot=False, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.Power} KW. The price of the component using Couper Method : {CI} euro")

            return float(CI)

        # Calculate cost of turbines
        elif self.Component_category == "Turbine":

            HP = self.Power * cF
            CI = a * (HP**n) * iF * ccF * 1000
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.Power} KW. The price of the component using Couper Method : {CI} euro ")

            return float(CI)

        # Calculate cost of shell and tube heat exchangers
        elif self.Component_category == "Heat Exchanger S/T":

            Area_sqft = self.Area * cF

            if "Shell and Tube, Fixed Head" in self.Component_name:
                fd = np.exp(-1.1156 + 0.0906 * np.log(Area_sqft))
            elif "Shell and Tube, Kettle Reboiler" in self.Component_name:
                fd = 1.35
            elif "Shell and Tube, U Tube" in self.Component_name:
                fd = np.exp(-0.9816 + 0.0830 * np.log(Area_sqft))

            if 689.476 <= self.Pressure <= 2068.43:
                fp = 0.7771 + 0.04981 * np.log(Area_sqft)
            elif 2068.43 < self.Pressure <= 4136.85:
                fp = 1.0305 + 0.07140 * np.log(Area_sqft)
            elif 4136.85 < self.Pressure <= 6205.28:
                fp = 1.1400 + 0.12088 * np.log(Area_sqft)
            else:
                print(
                    "The operational pressure for the shell and tube heat exchanger is out of range for the Couper method")

            Fm = b1 + b2 * np.log(Area_sqft)
            Cb = np.exp(8.821 - 0.30863 * np.log(Area_sqft) +
                        0.0681 * (np.log(Area_sqft) ** n))
            CI = a * fd * Fm * fp * Cb * ccF * iF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.Pressure} KPa and {self.Area} m^2. The price of the component using Couper Method : {CI} euro ")

            return float(CI)

        # Calculate cost of double pipe heat exchnagers
        elif self.Component_category == "Heat Exchanger D/P":

            Area_sqft = self.Area * cF

            if self.Pressure <= 400:
                fp = 1
            elif 400 < self.Pressure <= 600:
                fp = 1.10
            elif 600 < self.Pressure <= 700:
                fp = 1.25
            else:
                print(
                    "The operational pressure for the double pipe heat exchanger is out of range for the Couper method")

            CI = a * fm * fp * Area_sqft**n
            CI = CI * iF * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.Pressure} KPa and {self.Area} m^2. The price of the component using Couper Method : {CI} euro ")

            return float(CI)

        # Calculate cost of storage tanks
        elif self.Component_category == "Shop Fabricated Storage Tank":

            Volume_gal = self.Volume * cF
            CI = a * fm * \
                np.exp(2.631 + 1.3673 * np.log(Volume_gal) -
                       0.06309 * (np.log(Volume_gal))**n)
            CI = CI * iF * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.Volume} m^3. The price of the component using Couper Method : {CI} euro")

            return float(CI)

        # Calculate cost of centrifugal pumps
        elif self.Component_category == "VSC Centrifugal Pump" or self.Component_category == "HSC Centrifugal Pump":

            Q_gpm = self.Q * cF
            h_ft = self.h * cF2
            Cb = 3.00 * np.exp(8.833 - 0.6019 * np.log(Q_gpm * np.sqrt(h_ft)
                                                       ) + 0.0519 * (np.log(Q_gpm * np.sqrt(h_ft)))**n)
            ft = np.exp(b1 + b2 * np.log(Q_gpm * np.sqrt(h_ft)) +
                        b3 * np.log(Q_gpm * np.sqrt(h_ft))**n)
            CI = fm * ft * Cb * ccF * iF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.Q} m^3/s and {self.h} m. The price of the component using Couper Method : {CI} euro ")

            return float(CI)

        # Calculate cost of recipocating pumps
        elif self.Component_category == "Pump":
            print(
                "The costs produced for Reciprocating pumps using Couper Methods are unexpectedly high")
            Q_gpm = self.Q * cF
            CI = a * fm * (Q_gpm**n)
            CI = CI * iF * ccF * 1000
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.Q} m^3/s. The price of the component using Couper Method : {CI} euro ")

            return float(CI)

    def Towler_Method(self, Comp_dict, verbose=False, Desired_year=None):
        """
        Cost estimation according to:

        Towler and Gavin (2012)
        "Chemical engineering design: principles, practice and economics of plant and process design".
        
        SI-unit input is implemented! Limits are checked! BA 2025-08-15

        Equation
        ----------
            Ce = a + b * S**n.

        Parameters
        ----------
        Comp_dict : TYPE, Dictionary.
            DESCRIPTION : This dictionary contains all the variables and paramters needed to
            find a cost estimate for a particular component. For the exact values in the dictionary
            check the excel file Capital_Investment_Methods or check the textbook corresponding to this
            method.
            Comp_dict_example= {
                "Category": "Compressor",
                "Component Name": "hydrogen compressor",
                "Component Attribute" : 1000
                }
        CI : TYPE, Float.
        DESCRIPTION : capital investment of investigated component
        S : TYPE,  Float.
        DESCRIPTION : capacity of investigated component
        n : TYPE, Float.
        DESCRIPTION : cost exponent (n < 1 - economy of scales)
        a,b : TYPE, Float.
        DESCRIPTION : constant values for known device costs (literature)
        verbose : Boolean, optional
            DESCRIPTION : Should values be printed along with an explanation?
        Desired_year : TYPE, Float, optional.
            DESCRIPTION : The default is None. Do you want the cost index to be taken relative to the current year or another previous years?
            If you want the cost index to be takes relative to a previuos year, input value to Desired_year.


        Raises
        ------
        ValueError.
            DESCRIPTION : Checks if the component name is available in the Towler method excel sheet.

        Returns
        -------
        TYPE, String.
            DESCRIPTION : The installed cost of the component, adjusted to present day to account
            for inflation. This cost estimate also accounts for operational pressure and material of construction.

        """

        self.Component_category = Comp_dict.get("Category")
        self.Component_name = Comp_dict.get("Component Name")
        self.S = Comp_dict.get("Component Attribute")

        # Finds row corresponding to component and checks if component is available in this method
        index = np.where(self.Towler_xl_data[:, 1] == self.Component_name)[0]

        if index.size == 0:
            raise ValueError(
                "Component name is not found in the Towler method")

        # Array index contains the row for the component
        # Finds values needed for the cost estimation for the component
        Component_row = self.Towler_xl_data[index[0]]
        if Component_row[self.Towler_unit_col_index] in ("kW", "KW", "kPa", "KPa"):
            self.S = self.S/1000  # SI units, BA
        elif Component_row[self.Towler_unit_col_index] in ("l/s", "L/s"):
            self.S = self.S * 1000  # SI units, BA
        #limits = self.get_values_by_name(self.Component_name)
        self.check_value_within_limits(self.Component_name, self.S)

        a = float(Component_row[2])
        b = float(Component_row[3])
        n = float(Component_row[4])
        yr = float(Component_row[6])
        iF = float(Component_row[7])
        fm = float(Component_row[8])
        fl = float(Component_row[9])
        cF = float(Component_row[10])

        # Calculates cost of compoennt
        CI = a + b * (np.abs(self.S)**n)
        CI = self._Find_Present_Cost(CI, yr, plot=verbose, Current_year=(
            Desired_year is None), Desired_year=Desired_year) * iF * fm * fl * cF

        if verbose == True:
            print(
                f"The name of this component is {self.Component_name}. The component attriute is {self.S} (units). The price of the component using Towler Method : {CI} euro ")

        return float(CI)

    def get_values_by_name(self, name_value, columns=None, df=None):
        """
        Retrieve the values of specific columns for the row matching a given 'name' value.

        Parameters
        ----------
        name_value : str
            The value to search for in the 'name' column.
        columns : list of str, optional
            List of column names whose values should be returned.
            Defaults to ["S_lower", "S_upper"] if not provided.
        df : pandas.DataFrame, optional
            DataFrame in which to search. Defaults to self.Towler_df if not provided.

        Returns
        -------
        dict or None
            A dictionary mapping each requested column name to its value for the matching row,
            or None if no matching row is found.

        Notes
        -----
        - If multiple rows have the same 'name', only the first match is returned.
        - This method is flexible and can be used with any DataFrame and column list.
        """
        if df is None:
            df = self.Towler_df
        if columns is None:
            columns = ["S_lower", "S_upper"]

        row = df.loc[df["name"] == name_value]
        if row.empty:
            return None
        return {col: row.iloc[0][col] for col in columns}

    import warnings

    def check_value_within_limits(self, name_value, value,
                                  lower_col="S_lower", upper_col="S_upper",
                                  df=None, error=False):
        """
        Check if a given value lies within predefined limits from a DataFrame entry.

        Parameters
        ----------
        name_value : str
            The value to search for in the 'name' column of the DataFrame.
        value : float
            The numeric value to be checked.
        lower_col : str, optional
            Column name for the lower limit. Default is "S_lower".
        upper_col : str, optional
            Column name for the upper limit. Default is "S_upper".
        df : pandas.DataFrame, optional
            DataFrame to search in. Defaults to self.Towler_df.
        error : bool, optional
            If True, raises a ValueError when the value is out of range.
            If False, issues a UserWarning instead.

        Returns
        -------
        bool
            True if the value is within limits or if no limits are defined.
            False if the value is out of range.

        Notes
        -----
        - If the limits are missing or NaN, the check passes automatically.
        - If multiple rows match `name_value`, only the first is used.
        """
        self.warn = {'value': 0,
                     'message': 'All o.k.'}
        if df is None:
            df = self.Towler_df

        limits = self.get_values_by_name(
            name_value, columns=[lower_col, upper_col], df=df)

        if limits is None:
            return True  # No matching row → OK

        lower = limits.get(lower_col, None)
        upper = limits.get(upper_col, None)

        # If limits are missing or NaN → treat as OK
        if lower is None or upper is None or np.isnan(lower) or np.isnan(upper):
            return True

        if not (lower <= value <= upper):
            msg = (f"Value {value} is outside allowed range [{lower}, {upper}] "
                   f"for component '{name_value}'. The returned value is an extrapolation")
            self.warn = {'value': value/(lower + 1),
                         'message': msg}
            if error:
                raise ValueError(msg)
            else:
                with warnings.catch_warnings():
                    warnings.simplefilter("always")  # Oder "default"
                    warnings.warn(msg, UserWarning)
            return False

        return True

    def _Turton_Method(self, Comp_dict, verbose=False, Desired_year=None):
        """
        Cost estimation according to:

        Turton, R. (2003). Analysis, synthesis, and design of Chemical Processes. Prentice Hall.
        
        SI-unit input not implemented yet! Limits are not checked! BA 2025-08-15

        Parameters
        ----------
        Comp_dict : TYPE, Dictionary.
            DESCRIPTION. Contains input parameters which are used by the Turton method to calculate
            the cost of the  component. A sample of this dictionary would look like :
                Comp_dict = {
                "Category": "Pump",
                "Component Name": "Centrifugal Pump, Cast iron",
                "Component Attribute": 200,
                "Pressure": 1000             # For the turton method the pressures are in guage, 100 KPa = 1atm
                    }

        verbose : TYPE, Boolean, optional.
            should values be printed along with an explanation?
        Desired_year : TYPE, Float, optional
            Do you want the cost index to be taken relative to the current year or another previous years?

        Raises
        ------
        ValueError.
            DESCRIPTION : Checks if the component name is available in the Turton method excel sheet.

        Returns
        -------
        TYPE, String.
            DESCRIPTION : The cost of the component, adjusted to present day to account
            for inflation. This cost estimate also accounts for operational pressure and material of construction.

        """

        self.Component_category = Comp_dict.get("Component Catgeory")
        self.Component_name = Comp_dict.get("Component Name")
        self.S = Comp_dict.get("Component Attribute")
        self.Pressure = Comp_dict.get("Pressure")

        # Finds row corresponding to component and checks if component is available in this method
        index = np.where(self.Turton_xl_data[:, 1] == self.Component_name)[0]
        if index.size == 0:
            raise ValueError(
                "Component name is not found in the Turton method")

        # Array index contains the row for the component
        # Finds values needed for the cost estimation for the component
        Component_row = self.Turton_xl_data[index[0]]
        K1 = float(Component_row[2])
        K2 = float(Component_row[3])
        K3 = float(Component_row[4])
        B1 = float(Component_row[5])
        B2 = float(Component_row[6])
        n = float(Component_row[7])
        fm = float(Component_row[8])
        Fbm = float(Component_row[9])
        ccF = float(Component_row[10])
        cF2 = float(Component_row[18])
        yr = float(Component_row[19])
        fp = None
        c1 = None
        c2 = None
        c3 = None

        # Calculates the base cost of components, assuming ambient pressure and cs construction
        Cp_0 = K1 + K2 * np.log10(self.S) + K3 * (np.log10(self.S))**n
        Cp_0 = 10**Cp_0

        if any(x in self.Component_name for x in ["Compressor", "Turbine", "Liquid Expanders"]):

            CI = Cp_0 * Fbm * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.S} (units) and {self.Pressure} KPa. The price of the component using Turton Method : {CI} euro ")

            return float(CI)

        elif any(x in self.Component_name for x in ["Fixed Tube", "Floating Head", "U Tube"]):

            if self.Pressure <= 500:
                fp = 1
            elif 500 < self.Pressure <= 14000:
                c1 = -0.00164
                c2 = -0.00627
                c3 = -0.0123
                fp = c1 + c2 * np.log10(self.Pressure*cF2) + \
                    c3 * (np.log10(self.Pressure*cF2))**n
                fp = 10**fp
            else:
                print(
                    "The operational pressure for the shell and tube heat exchanger is out of range for the Turton method")

            CI = Cp_0 * (B1 + (B2*fm*fp))
            CI = CI * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.S} (units) and {self.Pressure} KPa. The price of the component using Turton Method : {CI} euro ")

            return float(CI)

        elif "Double Pipe" in self.Component_name:

            if self.Pressure <= 4000:
                fp = 1
            elif 4000 < self.Pressure <= 10000:
                c1 = 0.6072
                c2 = -0.9120
                c3 = 0.3327
                fp = c1 + c2 * np.log10(self.Pressure*cF2) + \
                    c3 * (np.log10(self.Pressure*cF2))**n
                fp = 10**fp
            else:
                print(
                    "The operational pressure for the shell and tube heat exchanger is out of range for the Turton method")

            CI = Cp_0 * (B1 + (B2*fm*fp))
            CI = CI * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.S} (units) and {self.Pressure} KPa. The price of the component using Turton Method : {CI} euro ")

            return float(CI)

        elif "Flat Plate" in self.Component_name:

            if self.Pressure < 1900:
                fp = 1
            else:
                print(
                    "The operational pressure for the shell and tube heat exchanger is out of range for the Turton method")
            CI = Cp_0 * (B1 + (B2*fm*fp))
            CI = CI * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.S} (units) and {self.Pressure} KPa. The price of the component using Turton Method : {CI} euro ")

            return float(CI)

        elif "Centrifugal Pump" in self.Component_name:

            if self.Pressure <= 1000:
                fp = 1
            elif 1000 < self.Pressure <= 10000:
                c1 = -0.3935
                c2 = -0.3957
                c3 = -0.00226
                fp = c1 + c2 * np.log10(self.Pressure*cF2) + \
                    c3 * (np.log10(self.Pressure*cF2))**n
                fp = 10**fp
            else:
                print(
                    "The operational pressure for the shell and tube heat exchanger is out of range for the Turton method")

            CI = Cp_0 * (B1 + (B2*fm*fp))
            CI = CI * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.S} (units) and {self.Pressure} KPa. The price of the component using Turton Method : {CI} euro ")

            return float(CI)

        elif any(x in self.Component_name for x in ["Positive Displacement Pump", "Reciprocating Pump"]):

            if self.Pressure <= 1000:
                fp = 1
            elif 1000 < self.Pressure <= 10000:
                c1 = -0.245382
                c2 = 0.259016
                c3 = -0.01363
                fp = c1 + c2 * np.log10(self.Pressure*cF2) + \
                    c3 * (np.log10(self.Pressure*cF2))**n
                fp = 10**fp
            else:
                print(
                    "The operational pressure for the shell and tube heat exchanger is out of range for the Turton method")

            CI = Cp_0 * (B1 + (B2*fm*fp))
            CI = CI * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.S} (units) and {self.Pressure} KPa. The price of the component using Turton Method : {CI} euro ")

            return float(CI)

        elif any(x in self.Component_name for x in ["Floating Roof", "Fixed Roof"]):

            if self.Pressure <= 7:
                fp = 1
            else:
                print(
                    "The operational pressure for the shell and tube heat exchanger is out of range for the Turton method")

            CI = Cp_0 * (B1 + (B2*fm*fp))
            CI = CI * ccF
            CI = self._Find_Present_Cost(CI, yr, plot=True, Current_year=(
                Desired_year is None), Desired_year=Desired_year)

            if verbose == True:
                print(
                    f"The name of this component is {self.Component_name}. The component attriute is {self.S} (units) and {self.Pressure} KPa. The price of the component using Turton Method : {CI} euro ")

            return float(CI)


"==========================================  TEST RUN.   =============================================="

if __name__ == "__main__":

    Eco = CAP_methods()

    Comp_dict = {
        "Category": "Compressor",
        "Component Name": "centrifugal compressor cs",
        "Component Attribute": 76000,  # SI units, Watt!

    }

    Cost_comp_couper_method = Eco.Towler_Method(
        Comp_dict, verbose=False, Desired_year=2024)
    print(f'Costs: {Cost_comp_couper_method:.3e}', Comp_dict, Eco.warn)



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 26 10:57:15 2024
wrapper to call ht-database and pressure loss (fluids.friction)

@author: welp
"""

import ht as ht
import carbatpy as cb
import fluids.friction as frict
import numpy as np
from fluids.two_phase import two_phase_dP



def alpha_1P(fluid, fluidstate, mdot, l, d_big, d_small=0):
    """
    heat transfer coefficient in 1phase flow from

    VDI-Wärmeatlas, Kap.G1, 11.Aufl., Springer Vieweg, 2013
    ISBN:  978-3-642-19980-6 / 978-3-642-19981-3
    
    calculates pressure loss with Darcy equiation and Colebrook quation for 
    Darcy friction factor
    

    Parameters
    ----------
    fluid : TYPE
        DESCRIPTION.
    fluidstate : TYPE
        DESCRIPTION. of fluid at given point
    mdot : TYPE
        DESCRIPTION. mass flow rate [kg/s]
    l : TYPE
        DESCRIPTION. length of tube (only used for flow field as reference 
                                     length) [m]
    d_big : TYPE
        DESCRIPTION. diameter [m]
    d_small : TYPE, optional
        DESCRIPTION. The default is 0. inner diameter if annulus [m]

    Returns
    -------
    alpha_i : TYPE
        DESCRIPTION. heat transfer coefficient [W/(m^2*K)]
    dp : TYPE
        DESCRIPTION. differential pressure loss [Pa/m]

    """

    Vdot = mdot * fluidstate.sp_volume  # V = m * v
    u = Vdot / (((d_big/2)**2 - (d_small/2)**2) * np.pi)  # [m/s]  u = Vdot / (pi * r^2)

    dh = d_big - d_small  # if inner tube, d_small zero

    Re = u * dh / (fluidstate.viscosity * fluidstate.sp_volume)  # B1-(10)   u * d / (eta * v)
    
    if Re > 10000:
        ep = 0.045*1e-3
        f_s = frict.Colebrook(Re, ep/dh)
        Xi = (1.8 * np.log10(Re) - 1.5)**-2  # G1-(27)

        Nu = ((Xi/8) * Re * fluidstate.prandtl)/(1 + 12.7 * (Xi/8)**0.5
                                 * (fluidstate.prandtl**(2/3) - 1))\
            * (1+(dh/l)**(2/3))  # G1-(26)

    elif Re > 2300:
        ep = 0.045*1e-3
        f_s = frict.Colebrook(Re, ep/dh)
        gamma = (Re - 2300)/(1e4 - 2300)  # G1-(30)

        Nu_wmq2 = 1.615 * (2300 * fluidstate.prandtl * dh/l)**(1/3)  # G1-(32)
        Nu_wmq3 = (2/(1+22 * fluidstate.prandtl))**(1/6) * (2300 * dh/l)**0.5  # G1-(33)
        Nu_wL = (49.371 + (Nu_wmq2-0.7)**3 + Nu_wmq3**3)**(1/3)  # G1-(31)

        Nu_wT = ((0.0308/8) * 1e4 * fluidstate.prandtl) / \
                (1 + 12.7 * (0.0308/8)**0.5 * (fluidstate.prandtl**(2/3) - 1))  # G1-(37)

        Nu = (1-gamma)*Nu_wL + gamma * Nu_wT  # G1-(29)

    else:
        Nu = 3.66  # G1-(1)
        f_s = 64/Re

    alpha_i = Nu * fluidstate.thermal_conductivity / dh  # [W/m²K] B1-(9)

    dp = f_s / fluidstate.sp_volume * 0.5 * (1/dh) * u**2   # Darcy equation Pa/m
    if np.iscomplexobj(alpha_i):
        raise ValueError(f"error in alpha_i: complex value detected")
    return float(alpha_i), dp

def alpha_km(fluid, fluidstate, mdot, method, l, d_big, U_i=0):
    """
    heat transfer coefficient and pressure drop for working fluid, changes
    automatically to 1phase if necessary.

    Parameters
    ----------
    fluid : TYPE
        DESCRIPTION.
    fluidstate : TYPE
        DESCRIPTION.
    mdot : TYPE
        DESCRIPTION. mass flow rate [kg/s]
    method : TYPE
        DESCRIPTION. Nusselt correlation as in ht database or constant
    l : TYPE
        DESCRIPTION. length of tube (only used for flow field as reference 
                                     length) [m]
    d_big : TYPE
        DESCRIPTION. diameter [m]
        
    U_i : TYPE, optional
        DESCRIPTION. The default is 0. please set, if constant U shall be used

    Returns
    -------
    alpha_KM : TYPE
        DESCRIPTION. heat transfer coefficient [W/(m^2*K)]
    dp_wf : TYPE
        DESCRIPTION. differential pressure loss [Pa/m]

    """  
    
    if 0 < fluidstate.quality < 1:
        # for 2phase
        fluidstate_vap = fluid.set_state([fluidstate.pressure, 1], "PQ", cb.fprop._TRANS_STRING, output="FluidState")
        
        fluidstate_liq = fluid.set_state([fluidstate.pressure, 0], "PQ", cb.fprop._TRANS_STRING, output="FluidState")  # TO DO verify if this saturated state
          # is how it is meant in publication
        
        dp_wf = two_phase_dP(mdot, fluidstate.quality, float(1/fluidstate_liq.sp_volume), d_big, 1,
                                 float(1/fluidstate_vap.sp_volume), float(fluidstate_liq.viscosity),
                                 float(fluidstate_vap.viscosity))
    # =============================================================================
    #         if method == "Chen_Bennett":
    # 
    #             alpha_KM = ht.boiling_flow.Chen_Bennett(
    #                 m, x, D, rhol, rhog, mul, mug, kl, Cpl, Hvap, sigma, dPsat, Te)
    # 
    #         elif method == "Chen_Edelstein":
    # 
    #             alpha_KM = ht.boiling_flow.Chen_Edelstein(
    #                 m, x, D, rhol, rhog, mul, mug, kl, Cpl, Hvap, sigma, dPsat, Te)
    # =============================================================================
            
        if method == "Cavallini_Smith_Zecchin":
            # heat transfer coefficient for condensation of a fluid inside a tube
            alpha_KM = ht.condensation.Cavallini_Smith_Zecchin(mdot, 
                                                               fluidstate.quality, 
                                                               d_big, 
                                                               1/fluidstate_liq.sp_volume, 
                                                               1/fluidstate_vap.sp_volume, 
                                                               fluidstate_liq.viscosity,
                                                               fluidstate_vap.viscosity,
                                                               fluidstate_liq.thermal_conductivity,
                                                               fluidstate_liq.cp,
                                                               )
        elif method == "Chen_Bennett":
            # heat transfer coefficient for film boiling of saturated fluid in any orientation of flow
            sigma = fluid.set_state([fluidstate.pressure, 0], "PQ", "STN")
            alpha_KM = ht.boiling_flow.Chen_Bennett(mdot,
                                                    fluidstate.quality,
                                                    d_big,
                                                    1/fluidstate_liq.sp_volume,
                                                    1/fluidstate_vap.sp_volume,
                                                    fluidstate_liq.viscosity,
                                                    fluidstate_vap.viscosity,
                                                    fluidstate_liq.thermal_conductivity,
                                                    fluidstate_liq.cp,
                                                    fluidstate_vap.enthalpy - fluidstate_liq.enthalpy,
                                                    sigma,
                                                    0,
                                                    0)      # Te and dPsat set to 0
            
        
        elif method == "constant":
            if U_i == 0:
                raise ValueError("U_i needs to be assigned if constant heat\
                                 transfer coefficient is used!")
            alpha_KM = U_i
                
        else:
            raise ValueError(f"method {method} for heat transfer not implemented")

    else:
        # if refrigerant in 1phasestate, mdot, l, d_big, d_small=0
        fluidstate = fluid.set_state([fluidstate.temperature, fluidstate.pressure], 
                                     "TP", cb.fprop._TRANS_STRING, output="FluidState")
        alpha_KM, dp_wf = alpha_1P(fluid, fluidstate, mdot, l, d_big)
    if np.iscomplexobj(alpha_KM):
        raise ValueError("error in alpha_wf: complex value detected")
    return float(alpha_KM) , dp_wf

if __name__ == "__main__":
    FLUID1 = "Propane * Pentane"
    comp1 = [.5, .5]
    fluid1= cb.fprop.init_fluid(FLUID1, comp1)
    st1_in = fluid1.set_state([350., 0.5], "TQ", output="FluidState")
    
    FLUID2 = "water"
    comp2 = [1]
    fluid2 = cb.fprop.init_fluid(FLUID2, comp2)
    st2_in =  fluid2.set_state([300., 1e5], "TP", cb.fprop._TRANS_STRING, output="FluidState")
    
    U_i = 250
    A = 8
    mdot_1, mdot_2 = 0.2, 0.4
    
    alpha1 = alpha_1P(fluid2, st2_in, mdot_2, 10, 20e-3, 15e-3)
    alpha2 = alpha_km(fluid1, st1_in, mdot_1, "Cavallini_Smith_Zecchin", 10, 15e-3)
    alpha3 = alpha_km(fluid1, st1_in, mdot_1, "Chen_Bennett", 10, 15e-3)
    alpha4 = alpha_km(fluid1, st1_in, mdot_1, "constant", 10, 15e-3, U_i = 100)


